#ifndef CPM_COLL_H_
#define CPM_COLL_H_

/*!
 * \page cpm_coll Model-based collectives
 * Shared library \c libcpm_coll.so implements different model-based algorithms of collectives:
 * - \ref lininterp_collectives
 * - \ref hockney_collectives
 * - \ref loggp_collectives
 * - \ref plogp_collectives
 * - \ref lmo_collectives
 *
 * Command-line arguments:
 * - \b verbose verbose mode (default: no)
 * - \b sgv scatterv/gatherv mode
 *   - 0 propagation (default)
 *   - 1 broadcast
 *   - 2 allover
 * - \b model \e S communication model: Hockney, PLogP, LMO (required for model-based collectives)
 * - \b file \e S model data file (required for model-based collectives)
 *
 * In order to preserve the original MPI interface, all model-based implementations use
 * the global variables that provide model parameters.
 * The interface of the implementation of a collective communication operation \c Y
 * based on the model \c X includes the following components:
 * - <code>int X_Y(standard args)</code> is the model-based collective operation itself
 * (for example, \ref Hockney_Scatter_bfs_binomial_min)
 * - <code>X_model* X_model_instance</code> is a global variable providing the model parameters
 * (must be available at all processors in the MPI communicator)
 * - <code>int X_initialize(MPI_Comm, X_model* model), int X_finalize(MPI_Comm)</code> are functions
 * responsible for allocation and deallocation of the model instance at all processors.
 * The \c model argument encapsulates the model parameters obtained either from a file or
 * the model builder.
 *
 * \dotfile collectives/collectives.dot
 *
 * All model-based algorithms are divided into two groups: model-specific and generic.
 *
 * <strong>Model-specific collectives</strong> depend on certain communication performance models,
 * using parameters specific for these models only. For example, \ref LMO_Gather_split_flat
 * directly uses the LMO model global variable and its threshold parameters
 * to split the medium size messages and perform a series of linear gathers with small messages,
 * in order to avoid escalations of the execution time on the clusters
 * with the TCP/IP communication layer. A model-specific collective operation \c Y is implemented
 * in the following way:
 * \code
int X_Y(standard args) {
	if (condition with X_model_instance->param)
		return ...;
}
 * \endcode
 *
 * <strong>Generic collectives</strong> depend on the prediction of the execution time of
 * some communication operation (communication primitive); they are parameterized by predictions,
 * which can be provided by any model. The communication primitive can be either
 * the collective operation itself or some other simple operation. See \ref cpm_collectives.
 *
 * \section tools Tools
 * Part of <a href="http://hcl.ucd.ie/project/mpiblib/api">MPIBlib</a>
 * - MPIBlib/tools/collective_test - performs a universal or operation-specific collective benchmark with given accuracy and efficiency
 * - MPIBlib/tools/collective - verifies implementations of collective communication operations
 *
 * \section tests Tests
 * - \subpage mod_coll
 * - \subpage pgemm
 */

#include "lininterp_collectives.h"
#include "hockney_collectives.h"
#include "loggp_collectives.h"
#include "plogp_collectives.h"
#include "lmo_collectives.h"

#endif /* CPM_COLL_H_ */
