#include "config.h"
#include "hockney_collectives.h"
#include "cpm_collectives.h"
#include <memory.h>

/*!
 * The global instance of Hockney model for use in the optimized scatter/gather.
 * Must be initialized by \ref Hockney_initialize and destroyed by \ref Hockney_finalize
 * at all processes.
 */
Hockney_model* Hockney_model_instance = NULL;

int Hockney_initialize(MPI_Comm comm, Hockney_model* model) {
	int rank;
	MPI_Comm_rank(comm, &rank);
	int n = rank == 0 ? model->n : 0;
	MPI_Bcast(&n, 1, MPI_INT, 0, comm); 
	Hockney_model_instance = Hockney_alloc(n);
	int Cn2 = n * (n - 1) / 2;
	if (rank == 0) {
		memcpy(Hockney_model_instance->a, model->a, sizeof(double) * Cn2);
		memcpy(Hockney_model_instance->b, model->b, sizeof(double) * Cn2);
	}
	MPI_Datatype type;
	int blocklens[2] = {Cn2, Cn2} ;
	MPI_Aint displs[2] = {
			(void*)Hockney_model_instance->a - (void*)Hockney_model_instance,
			(void*)Hockney_model_instance->b - (void*)Hockney_model_instance
	};
	MPI_Datatype types[2] = {MPI_DOUBLE, MPI_DOUBLE};
	MPI_Type_struct(2, blocklens, displs, types, &type);
	MPI_Type_commit(&type);
	MPI_Bcast(Hockney_model_instance, 1, type, 0, comm);
	MPI_Type_free(&type);
	return 0;
}

int Hockney_finalize(MPI_Comm comm) {
	Hockney_free(Hockney_model_instance);
	Hockney_model_instance = NULL;
	return 0;
}

int Hockney_Scatterv_sorted_flat_asc(void* sendbuf, int* sendcounts, int* displs, MPI_Datatype sendtype,
	void* recvbuf, int recvcount, MPI_Datatype recvtype,
	int root, MPI_Comm comm)
{
	if (Hockney_model_instance == NULL)
		return CPM_ERR_MODEL;
	return CPM_Scatterv_sorted_flat(&Hockney_model_instance->predictor, ASC,
		sendbuf, sendcounts, displs, sendtype, recvbuf, recvcount, recvtype, root, comm);
}

int Hockney_Gatherv_sorted_flat_asc(void* sendbuf, int sendcount, MPI_Datatype sendtype,
	void* recvbuf, int* recvcounts, int* displs, MPI_Datatype recvtype,
	int root, MPI_Comm comm)
{
	if (Hockney_model_instance == NULL)
		return CPM_ERR_MODEL;
	return CPM_Gatherv_sorted_flat(&Hockney_model_instance->predictor, ASC,
		sendbuf, sendcount, sendtype, recvbuf, recvcounts, displs, recvtype, root, comm);
}

int Hockney_Scatterv_sorted_flat_dsc(void* sendbuf, int* sendcounts, int* displs, MPI_Datatype sendtype,
	void* recvbuf, int recvcount, MPI_Datatype recvtype,
	int root, MPI_Comm comm)
{
	if (Hockney_model_instance == NULL)
		return CPM_ERR_MODEL;
	return CPM_Scatterv_sorted_flat(&Hockney_model_instance->predictor, DESC,
		sendbuf, sendcounts, displs, sendtype, recvbuf, recvcount, recvtype, root, comm);
}

int Hockney_Gatherv_sorted_flat_dsc(void* sendbuf, int sendcount, MPI_Datatype sendtype,
	void* recvbuf, int* recvcounts, int* displs, MPI_Datatype recvtype,
	int root, MPI_Comm comm)
{
	if (Hockney_model_instance == NULL)
		return CPM_ERR_MODEL;
	return CPM_Gatherv_sorted_flat(&Hockney_model_instance->predictor, DESC,
		sendbuf, sendcount, sendtype, recvbuf, recvcounts, displs, recvtype, root, comm);
}

int Hockney_Bcast_dfs_binomial_min(void* buffer, int count, MPI_Datatype datatype, int root, MPI_Comm comm) {
	if (Hockney_model_instance == NULL)
		return CPM_ERR_MODEL;
	return CPM_Bcast_dfs_binomial(&Hockney_model_instance->predictor, MIN,
		buffer, count, datatype, root, comm);
}

int Hockney_Bcast_dfs_binomial_max(void* buffer, int count, MPI_Datatype datatype, int root, MPI_Comm comm) {
	if (Hockney_model_instance == NULL)
		return CPM_ERR_MODEL;
	return CPM_Bcast_dfs_binomial(&Hockney_model_instance->predictor, MAX,
		buffer, count, datatype, root, comm);
}

int Hockney_Bcast_bfs_binomial_min(void* buffer, int count, MPI_Datatype datatype, int root, MPI_Comm comm) {
	if (Hockney_model_instance == NULL)
		return CPM_ERR_MODEL;
	return CPM_Bcast_bfs_binomial(&Hockney_model_instance->predictor, MIN,
		buffer, count, datatype, root, comm);
}

int Hockney_Bcast_bfs_binomial_max(void* buffer, int count, MPI_Datatype datatype, int root, MPI_Comm comm) {
	if (Hockney_model_instance == NULL)
		return CPM_ERR_MODEL;
	return CPM_Bcast_bfs_binomial(&Hockney_model_instance->predictor, MAX,
		buffer, count, datatype, root, comm);
}

int Hockney_Reduce_dfs_binomial_min(void* sendbuf, void* recvbuf, int count, MPI_Datatype datatype,
	MPI_Op op, int root, MPI_Comm comm)
{
	if (Hockney_model_instance == NULL)
		return CPM_ERR_MODEL;
	return CPM_Reduce_dfs_binomial(&Hockney_model_instance->predictor, MIN,
		sendbuf, recvbuf, count, datatype,
		op, root, comm);
}

int Hockney_Reduce_bfs_binomial_min(void* sendbuf, void* recvbuf, int count, MPI_Datatype datatype,
	MPI_Op op, int root, MPI_Comm comm) {
	if (Hockney_model_instance == NULL)
		return CPM_ERR_MODEL;
	return CPM_Reduce_bfs_binomial(&Hockney_model_instance->predictor, MIN,
		sendbuf, recvbuf, count, datatype,
		op, root, comm);
}

int Hockney_Reduce_bfs_binomial_max(void* sendbuf, void* recvbuf, int count, MPI_Datatype datatype,
	MPI_Op op, int root, MPI_Comm comm) {
	if (Hockney_model_instance == NULL)
		return CPM_ERR_MODEL;
	return CPM_Reduce_bfs_binomial(&Hockney_model_instance->predictor, MAX,
		sendbuf, recvbuf, count, datatype,
		op, root, comm);
}

int Hockney_Reduce_dfs_binomial_max(void* sendbuf, void* recvbuf, int count, MPI_Datatype datatype,
	MPI_Op op, int root, MPI_Comm comm)
{
	if (Hockney_model_instance == NULL)
		return CPM_ERR_MODEL;
	return CPM_Reduce_dfs_binomial(&Hockney_model_instance->predictor, MAX,
		sendbuf, recvbuf, count, datatype,
		op, root, comm);
}

int Hockney_Bcast_ucs_binomial_min(void* buffer, int count, MPI_Datatype datatype, int root, MPI_Comm comm) {
	if (Hockney_model_instance == NULL)
		return CPM_ERR_MODEL;
	return CPM_Bcast_ucs_binomial(&Hockney_model_instance->predictor, MIN,
		buffer, count, datatype, root, comm);
}

int Hockney_Reduce_ucs_binomial_min(void* sendbuf, void* recvbuf, int count, MPI_Datatype datatype,
	MPI_Op op, int root, MPI_Comm comm)
{
	if (Hockney_model_instance == NULL)
		return CPM_ERR_MODEL;
	return CPM_Reduce_ucs_binomial(&Hockney_model_instance->predictor, MIN,
		sendbuf, recvbuf, count, datatype,
		op, root, comm);
}

int Hockney_Bcast_ucs_binomial_max(void* buffer, int count, MPI_Datatype datatype, int root, MPI_Comm comm) {
	if (Hockney_model_instance == NULL)
		return CPM_ERR_MODEL;
	return CPM_Bcast_ucs_binomial(&Hockney_model_instance->predictor, MAX,
		buffer, count, datatype, root, comm);
}

int Hockney_Reduce_ucs_binomial_max(void* sendbuf, void* recvbuf, int count, MPI_Datatype datatype,
	MPI_Op op, int root, MPI_Comm comm)
{
	if (Hockney_model_instance == NULL)
		return CPM_ERR_MODEL;
	return CPM_Reduce_ucs_binomial(&Hockney_model_instance->predictor, MAX,
		sendbuf, recvbuf, count, datatype,
		op, root, comm);
}

int Hockney_Scatter_dfs_binomial_min(void* sendbuf, int sendcount, MPI_Datatype sendtype,
	void* recvbuf, int recvcount, MPI_Datatype recvtype, int root, MPI_Comm comm) {
	if (Hockney_model_instance == NULL)
		return CPM_ERR_MODEL;
	return CPM_Scatter_dfs_binomial(&Hockney_model_instance->predictor, MIN,
		sendbuf, sendcount, sendtype, recvbuf, recvcount, recvtype, root, comm);
}

/*! DFS binomial gather based on the Hockney model. */
int Hockney_Gather_dfs_binomial_min(void* sendbuf, int sendcount, MPI_Datatype sendtype,
	void* recvbuf, int recvcount, MPI_Datatype recvtype, int root, MPI_Comm comm) {
	if (Hockney_model_instance == NULL)
		return CPM_ERR_MODEL;
	return CPM_Gather_dfs_binomial(&Hockney_model_instance->predictor, MIN,
		sendbuf, sendcount, sendtype, recvbuf, recvcount, recvtype, root, comm);

}

/*! DFS binomial scatter based on the Hockney model. */
int Hockney_Scatter_dfs_binomial_max(void* sendbuf, int sendcount, MPI_Datatype sendtype,
	void* recvbuf, int recvcount, MPI_Datatype recvtype, int root, MPI_Comm comm) {
	if (Hockney_model_instance == NULL)
		return CPM_ERR_MODEL;
	return CPM_Scatter_dfs_binomial(&Hockney_model_instance->predictor, MAX,
		sendbuf, sendcount, sendtype, recvbuf, recvcount, recvtype, root, comm);
}


/*! DFS binomial gather based on the Hockney model. */
int Hockney_Gather_dfs_binomial_max(void* sendbuf, int sendcount, MPI_Datatype sendtype,
	void* recvbuf, int recvcount, MPI_Datatype recvtype, int root, MPI_Comm comm) {
	if (Hockney_model_instance == NULL)
		return CPM_ERR_MODEL;
	return CPM_Gather_dfs_binomial(&Hockney_model_instance->predictor, MAX,
		sendbuf, sendcount, sendtype, recvbuf, recvcount, recvtype, root, comm);
}

int Hockney_Scatter_ucs_binomial_min(void* sendbuf, int sendcount, MPI_Datatype sendtype,
	void* recvbuf, int recvcount, MPI_Datatype recvtype,
	int root, MPI_Comm comm)
{
	if (Hockney_model_instance == NULL)
		return CPM_ERR_MODEL;
	return CPM_Scatter_ucs_binomial(&Hockney_model_instance->predictor, MIN,
		sendbuf, sendcount, sendtype, recvbuf, recvcount, recvtype, root, comm);
}

int Hockney_Gather_ucs_binomial_min(void* sendbuf, int sendcount, MPI_Datatype sendtype,
	void* recvbuf, int recvcount, MPI_Datatype recvtype,
	int root, MPI_Comm comm)
{
	if (Hockney_model_instance == NULL)
		return CPM_ERR_MODEL;
	return CPM_Gather_ucs_binomial(&Hockney_model_instance->predictor, MIN,
		sendbuf, sendcount, sendtype, recvbuf, recvcount, recvtype, root, comm);
}

int Hockney_Scatter_ucs_binomial_max(void* sendbuf, int sendcount, MPI_Datatype sendtype,
	void* recvbuf, int recvcount, MPI_Datatype recvtype,
	int root, MPI_Comm comm)
{
	if (Hockney_model_instance == NULL)
		return CPM_ERR_MODEL;
	return CPM_Scatter_ucs_binomial(&Hockney_model_instance->predictor, MAX,
		sendbuf, sendcount, sendtype, recvbuf, recvcount, recvtype, root, comm);
}

int Hockney_Gather_ucs_binomial_max(void* sendbuf, int sendcount, MPI_Datatype sendtype,
	void* recvbuf, int recvcount, MPI_Datatype recvtype,
	int root, MPI_Comm comm)
{
	if (Hockney_model_instance == NULL)
		return CPM_ERR_MODEL;
	return CPM_Gather_ucs_binomial(&Hockney_model_instance->predictor, MAX,
		sendbuf, sendcount, sendtype, recvbuf, recvcount, recvtype, root, comm);
}

int Hockney_Scatter_bfs_binomial_min(void* sendbuf, int sendcount, MPI_Datatype sendtype,
	void* recvbuf, int recvcount, MPI_Datatype recvtype,
	int root, MPI_Comm comm)
{
	if (Hockney_model_instance == NULL)
		return CPM_ERR_MODEL;
	return CPM_Scatter_bfs_binomial(&Hockney_model_instance->predictor, MIN,
		sendbuf, sendcount, sendtype, recvbuf, recvcount, recvtype, root, comm);
}

int Hockney_Gather_bfs_binomial_min(void* sendbuf, int sendcount, MPI_Datatype sendtype,
	void* recvbuf, int recvcount, MPI_Datatype recvtype,
	int root, MPI_Comm comm)
{
	if (Hockney_model_instance == NULL)
		return CPM_ERR_MODEL;
	return CPM_Gather_bfs_binomial(&Hockney_model_instance->predictor, MIN,
		sendbuf, sendcount, sendtype, recvbuf, recvcount, recvtype, root, comm);
}

int Hockney_Scatter_bfs_binomial_max(void* sendbuf, int sendcount, MPI_Datatype sendtype,
	void* recvbuf, int recvcount, MPI_Datatype recvtype,
	int root, MPI_Comm comm)
{
	if (Hockney_model_instance == NULL)
		return CPM_ERR_MODEL;
	return CPM_Scatter_bfs_binomial(&Hockney_model_instance->predictor, MAX,
		sendbuf, sendcount, sendtype, recvbuf, recvcount, recvtype, root, comm);
}

int Hockney_Gather_bfs_binomial_max(void* sendbuf, int sendcount, MPI_Datatype sendtype,
	void* recvbuf, int recvcount, MPI_Datatype recvtype,
	int root, MPI_Comm comm)
{
	if (Hockney_model_instance == NULL)
		return CPM_ERR_MODEL;
	return CPM_Gather_bfs_binomial(&Hockney_model_instance->predictor, MAX,
		sendbuf, sendcount, sendtype, recvbuf, recvcount, recvtype, root, comm);
}

int Hockney_Scatterv_dfs_binomial_min(void* sendbuf, int* sendcounts, int* displs, MPI_Datatype sendtype,
	void* recvbuf, int recvcount, MPI_Datatype recvtype,
	int root, MPI_Comm comm)
{
	if (Hockney_model_instance == NULL)
		return CPM_ERR_MODEL;
	return CPM_Scatterv_dfs_binomial(&Hockney_model_instance->predictor, MIN,
		sendbuf, sendcounts, displs, sendtype, recvbuf, recvcount, recvtype, root, comm);
}

int Hockney_Gatherv_dfs_binomial_min(void* sendbuf, int sendcount, MPI_Datatype sendtype,
	void* recvbuf, int* recvcounts, int* displs, MPI_Datatype recvtype,
	int root, MPI_Comm comm)
{
	if (Hockney_model_instance == NULL)
		return CPM_ERR_MODEL;
	return CPM_Gatherv_dfs_binomial(&Hockney_model_instance->predictor, MIN,
		sendbuf, sendcount, sendtype, recvbuf, recvcounts, displs, recvtype, root, comm);
}

int Hockney_Scatterv_dfs_binomial_max(void* sendbuf, int* sendcounts, int* displs, MPI_Datatype sendtype,
	void* recvbuf, int recvcount, MPI_Datatype recvtype,
	int root, MPI_Comm comm)
{
	if (Hockney_model_instance == NULL)
		return CPM_ERR_MODEL;
	return CPM_Scatterv_dfs_binomial(&Hockney_model_instance->predictor, MAX,
		sendbuf, sendcounts, displs, sendtype, recvbuf, recvcount, recvtype, root, comm);
}

int Hockney_Gatherv_dfs_binomial_max(void* sendbuf, int sendcount, MPI_Datatype sendtype,
	void* recvbuf, int* recvcounts, int* displs, MPI_Datatype recvtype,
	int root, MPI_Comm comm)
{
	if (Hockney_model_instance == NULL)
		return CPM_ERR_MODEL;
	return CPM_Gatherv_dfs_binomial(&Hockney_model_instance->predictor, MAX,
		sendbuf, sendcount, sendtype, recvbuf, recvcounts, displs, recvtype, root, comm);
}

int Hockney_Scatterv_bfs_binomial_min(void* sendbuf, int* sendcounts, int* displs, MPI_Datatype sendtype,
	void* recvbuf, int recvcount, MPI_Datatype recvtype,
	int root, MPI_Comm comm)
{
	if (Hockney_model_instance == NULL)
		return CPM_ERR_MODEL;
	return CPM_Scatterv_bfs_binomial(&Hockney_model_instance->predictor, MIN,
		sendbuf, sendcounts, displs, sendtype, recvbuf, recvcount, recvtype, root, comm);
}

int Hockney_Gatherv_bfs_binomial_min(void* sendbuf, int sendcount, MPI_Datatype sendtype,
	void* recvbuf, int* recvcounts, int* displs, MPI_Datatype recvtype,
	int root, MPI_Comm comm)
{
	if (Hockney_model_instance == NULL)
		return CPM_ERR_MODEL;
	return CPM_Gatherv_bfs_binomial(&Hockney_model_instance->predictor, MIN,
		sendbuf, sendcount, sendtype, recvbuf, recvcounts, displs, recvtype, root, comm);
}

int Hockney_Scatterv_bfs_binomial_max(void* sendbuf, int* sendcounts, int* displs, MPI_Datatype sendtype,
	void* recvbuf, int recvcount, MPI_Datatype recvtype,
	int root, MPI_Comm comm)
{
	if (Hockney_model_instance == NULL)
		return CPM_ERR_MODEL;
	return CPM_Scatterv_bfs_binomial(&Hockney_model_instance->predictor, MAX,
		sendbuf, sendcounts, displs, sendtype, recvbuf, recvcount, recvtype, root, comm);
}

int Hockney_Gatherv_bfs_binomial_max(void* sendbuf, int sendcount, MPI_Datatype sendtype,
	void* recvbuf, int* recvcounts, int* displs, MPI_Datatype recvtype,
	int root, MPI_Comm comm)
{
	if (Hockney_model_instance == NULL)
		return CPM_ERR_MODEL;
	return CPM_Gatherv_bfs_binomial(&Hockney_model_instance->predictor, MAX,
		sendbuf, sendcount, sendtype, recvbuf, recvcounts, displs, recvtype, root, comm);
}

int Hockney_Scatterv_ucs_binomial_min(void* sendbuf, int* sendcounts, int* displs, MPI_Datatype sendtype,
	void* recvbuf, int recvcount, MPI_Datatype recvtype,
	int root, MPI_Comm comm)
{
	if (Hockney_model_instance == NULL)
		return CPM_ERR_MODEL;
	return CPM_Scatterv_ucs_binomial(&Hockney_model_instance->predictor, MIN,
		sendbuf, sendcounts, displs, sendtype, recvbuf, recvcount, recvtype, root, comm);
}

int Hockney_Gatherv_ucs_binomial_min(void* sendbuf, int sendcount, MPI_Datatype sendtype,
	void* recvbuf, int* recvcounts, int* displs, MPI_Datatype recvtype,
	int root, MPI_Comm comm)
{
	if (Hockney_model_instance == NULL)
		return CPM_ERR_MODEL;
	return CPM_Gatherv_ucs_binomial(&Hockney_model_instance->predictor, MIN,
		sendbuf, sendcount, sendtype, recvbuf, recvcounts, displs, recvtype, root, comm);
}

int Hockney_Scatterv_ucs_binomial_max(void* sendbuf, int* sendcounts, int* displs, MPI_Datatype sendtype,
	void* recvbuf, int recvcount, MPI_Datatype recvtype,
	int root, MPI_Comm comm)
{
	if (Hockney_model_instance == NULL)
		return CPM_ERR_MODEL;
	return CPM_Scatterv_ucs_binomial(&Hockney_model_instance->predictor, MAX,
		sendbuf, sendcounts, displs, sendtype, recvbuf, recvcount, recvtype, root, comm);
}

int Hockney_Gatherv_ucs_binomial_max(void* sendbuf, int sendcount, MPI_Datatype sendtype,
	void* recvbuf, int* recvcounts, int* displs, MPI_Datatype recvtype,
	int root, MPI_Comm comm)
{
	if (Hockney_model_instance == NULL)
		return CPM_ERR_MODEL;
	return CPM_Gatherv_ucs_binomial(&Hockney_model_instance->predictor, MAX,
		sendbuf, sendcount, sendtype, recvbuf, recvcounts, displs, recvtype, root, comm);
}

int Hockney_Scatterv_Traff(void* sendbuf, int* sendcounts, int* displs, MPI_Datatype sendtype,
	void* recvbuf, int recvcount, MPI_Datatype recvtype,
	int root, MPI_Comm comm)
{
	if (Hockney_model_instance == NULL)
		return CPM_ERR_MODEL;
	return CPM_Scatterv_Traff(&Hockney_model_instance->predictor,
		sendbuf, sendcounts, displs, sendtype,
		recvbuf, recvcount, recvtype,
		root, comm);
}

int Hockney_Gatherv_Traff(void* sendbuf, int sendcount, MPI_Datatype sendtype,
	void* recvbuf, int* recvcounts, int* displs, MPI_Datatype recvtype,
	int root, MPI_Comm comm)
{
	if (Hockney_model_instance == NULL)
		return CPM_ERR_MODEL;
	return CPM_Gatherv_Traff(&Hockney_model_instance->predictor,
		sendbuf, sendcount, sendtype,
		recvbuf, recvcounts, displs, recvtype,
		root, comm);
}

