#include "config.h"
#include "lininterp_collectives.h"
#include "cpm_collectives.h"
#include <memory.h>

/*!
 * The global instance of LinInterp model for use in the optimized scatter/gather.
 * Must be initialized by \ref LinInterp_initialize and destroyed by \ref LinInterp_finalize
 * at all processes.
 */
LinInterp_model* LinInterp_model_instance = NULL;

int LinInterp_initialize(MPI_Comm comm, LinInterp_model* model) {
	int rank;
	MPI_Comm_rank(comm, &rank);
	int n = rank == 0 ? model->n : 0;
	int num_points = rank == 0 ? model->num_points : 0;
	MPI_Bcast(&n, 1, MPI_INT, 0, comm); 
	MPI_Bcast(&num_points, 1, MPI_INT, 0, comm); 
	LinInterp_model_instance = LinInterp_alloc(n, num_points);
	double* x = (double*)malloc(sizeof(double) * num_points);
	double* y = (double*)malloc(sizeof(double) * num_points);
	int i;
	int Cn2 = n * (n - 1) / 2;
	for (i = 0; i < Cn2; i++) {
		if (rank == 0) {
			memcpy(x, model->splines[i]->x, sizeof(double) * num_points);
			memcpy(y, model->splines[i]->y, sizeof(double) * num_points);
		}
		MPI_Bcast(x, num_points, MPI_DOUBLE, 0, comm);
		MPI_Bcast(y, num_points, MPI_DOUBLE, 0, comm);
		gsl_spline_init(LinInterp_model_instance->splines[i], x, y, num_points);
	}
	free(x);
	free(y);
	return 0;
}

int LinInterp_finalize(MPI_Comm comm) {
	LinInterp_free(LinInterp_model_instance);
	LinInterp_model_instance = NULL;
	return 0;
}

int LinInterp_Scatterv_sorted_flat_asc(void* sendbuf, int* sendcounts, int* displs, MPI_Datatype sendtype,
	void* recvbuf, int recvcount, MPI_Datatype recvtype,
	int root, MPI_Comm comm)
{
	if (LinInterp_model_instance == NULL)
		return CPM_ERR_MODEL;
	return CPM_Scatterv_sorted_flat(&LinInterp_model_instance->predictor, ASC,
		sendbuf, sendcounts, displs, sendtype, recvbuf, recvcount, recvtype, root, comm);
}

int LinInterp_Gatherv_sorted_flat_asc(void* sendbuf, int sendcount, MPI_Datatype sendtype,
	void* recvbuf, int* recvcounts, int* displs, MPI_Datatype recvtype,
	int root, MPI_Comm comm)
{
	if (LinInterp_model_instance == NULL)
		return CPM_ERR_MODEL;
	return CPM_Gatherv_sorted_flat(&LinInterp_model_instance->predictor, ASC,
		sendbuf, sendcount, sendtype, recvbuf, recvcounts, displs, recvtype, root, comm);
}

int LinInterp_Scatterv_sorted_flat_dsc(void* sendbuf, int* sendcounts, int* displs, MPI_Datatype sendtype,
	void* recvbuf, int recvcount, MPI_Datatype recvtype,
	int root, MPI_Comm comm)
{
	if (LinInterp_model_instance == NULL)
		return CPM_ERR_MODEL;
	return CPM_Scatterv_sorted_flat(&LinInterp_model_instance->predictor, DESC,
		sendbuf, sendcounts, displs, sendtype, recvbuf, recvcount, recvtype, root, comm);
}

int LinInterp_Gatherv_sorted_flat_dsc(void* sendbuf, int sendcount, MPI_Datatype sendtype,
	void* recvbuf, int* recvcounts, int* displs, MPI_Datatype recvtype,
	int root, MPI_Comm comm)
{
	if (LinInterp_model_instance == NULL)
		return CPM_ERR_MODEL;
	return CPM_Gatherv_sorted_flat(&LinInterp_model_instance->predictor, DESC,
		sendbuf, sendcount, sendtype, recvbuf, recvcounts, displs, recvtype, root, comm);
}

int LinInterp_Bcast_dfs_binomial_min(void* buffer, int count, MPI_Datatype datatype, int root, MPI_Comm comm) {
	if (LinInterp_model_instance == NULL)
		return CPM_ERR_MODEL;
	return CPM_Bcast_dfs_binomial(&LinInterp_model_instance->predictor, MIN,
		buffer, count, datatype, root, comm);
}

int LinInterp_Bcast_dfs_binomial_max(void* buffer, int count, MPI_Datatype datatype, int root, MPI_Comm comm) {
	if (LinInterp_model_instance == NULL)
		return CPM_ERR_MODEL;
	return CPM_Bcast_dfs_binomial(&LinInterp_model_instance->predictor, MAX,
		buffer, count, datatype, root, comm);
}

int LinInterp_Bcast_bfs_binomial_min(void* buffer, int count, MPI_Datatype datatype, int root, MPI_Comm comm) {
	if (LinInterp_model_instance == NULL)
		return CPM_ERR_MODEL;
	return CPM_Bcast_bfs_binomial(&LinInterp_model_instance->predictor, MIN,
		buffer, count, datatype, root, comm);
}

int LinInterp_Bcast_bfs_binomial_max(void* buffer, int count, MPI_Datatype datatype, int root, MPI_Comm comm) {
	if (LinInterp_model_instance == NULL)
		return CPM_ERR_MODEL;
	return CPM_Bcast_bfs_binomial(&LinInterp_model_instance->predictor, MAX,
		buffer, count, datatype, root, comm);
}

int LinInterp_Reduce_dfs_binomial_min(void* sendbuf, void* recvbuf, int count, MPI_Datatype datatype,
	MPI_Op op, int root, MPI_Comm comm)
{
	if (LinInterp_model_instance == NULL)
		return CPM_ERR_MODEL;
	return CPM_Reduce_dfs_binomial(&LinInterp_model_instance->predictor, MIN,
		sendbuf, recvbuf, count, datatype,
		op, root, comm);
}

int LinInterp_Reduce_bfs_binomial_min(void* sendbuf, void* recvbuf, int count, MPI_Datatype datatype,
	MPI_Op op, int root, MPI_Comm comm) {
	if (LinInterp_model_instance == NULL)
		return CPM_ERR_MODEL;
	return CPM_Reduce_bfs_binomial(&LinInterp_model_instance->predictor, MIN,
		sendbuf, recvbuf, count, datatype,
		op, root, comm);
}

int LinInterp_Reduce_bfs_binomial_max(void* sendbuf, void* recvbuf, int count, MPI_Datatype datatype,
	MPI_Op op, int root, MPI_Comm comm) {
	if (LinInterp_model_instance == NULL)
		return CPM_ERR_MODEL;
	return CPM_Reduce_bfs_binomial(&LinInterp_model_instance->predictor, MAX,
		sendbuf, recvbuf, count, datatype,
		op, root, comm);
}

int LinInterp_Reduce_dfs_binomial_max(void* sendbuf, void* recvbuf, int count, MPI_Datatype datatype,
	MPI_Op op, int root, MPI_Comm comm)
{
	if (LinInterp_model_instance == NULL)
		return CPM_ERR_MODEL;
	return CPM_Reduce_dfs_binomial(&LinInterp_model_instance->predictor, MAX,
		sendbuf, recvbuf, count, datatype,
		op, root, comm);
}

int LinInterp_Bcast_ucs_binomial_min(void* buffer, int count, MPI_Datatype datatype, int root, MPI_Comm comm) {
	if (LinInterp_model_instance == NULL)
		return CPM_ERR_MODEL;
	return CPM_Bcast_ucs_binomial(&LinInterp_model_instance->predictor, MIN,
		buffer, count, datatype, root, comm);
}

int LinInterp_Reduce_ucs_binomial_min(void* sendbuf, void* recvbuf, int count, MPI_Datatype datatype,
	MPI_Op op, int root, MPI_Comm comm)
{
	if (LinInterp_model_instance == NULL)
		return CPM_ERR_MODEL;
	return CPM_Reduce_ucs_binomial(&LinInterp_model_instance->predictor, MIN,
		sendbuf, recvbuf, count, datatype,
		op, root, comm);
}

int LinInterp_Bcast_ucs_binomial_max(void* buffer, int count, MPI_Datatype datatype, int root, MPI_Comm comm) {
	if (LinInterp_model_instance == NULL)
		return CPM_ERR_MODEL;
	return CPM_Bcast_ucs_binomial(&LinInterp_model_instance->predictor, MAX,
		buffer, count, datatype, root, comm);
}

int LinInterp_Reduce_ucs_binomial_max(void* sendbuf, void* recvbuf, int count, MPI_Datatype datatype,
	MPI_Op op, int root, MPI_Comm comm)
{
	if (LinInterp_model_instance == NULL)
		return CPM_ERR_MODEL;
	return CPM_Reduce_ucs_binomial(&LinInterp_model_instance->predictor, MAX,
		sendbuf, recvbuf, count, datatype,
		op, root, comm);
}

int LinInterp_Scatter_dfs_binomial_min(void* sendbuf, int sendcount, MPI_Datatype sendtype,
	void* recvbuf, int recvcount, MPI_Datatype recvtype, int root, MPI_Comm comm) {
	if (LinInterp_model_instance == NULL)
		return CPM_ERR_MODEL;
	return CPM_Scatter_dfs_binomial(&LinInterp_model_instance->predictor, MIN,
		sendbuf, sendcount, sendtype, recvbuf, recvcount, recvtype, root, comm);
}

/*! DFS binomial gather based on the LinInterp model. */
int LinInterp_Gather_dfs_binomial_min(void* sendbuf, int sendcount, MPI_Datatype sendtype,
	void* recvbuf, int recvcount, MPI_Datatype recvtype, int root, MPI_Comm comm) {
	if (LinInterp_model_instance == NULL)
		return CPM_ERR_MODEL;
	return CPM_Gather_dfs_binomial(&LinInterp_model_instance->predictor, MIN,
		sendbuf, sendcount, sendtype, recvbuf, recvcount, recvtype, root, comm);

}

/*! DFS binomial scatter based on the LinInterp model. */
int LinInterp_Scatter_dfs_binomial_max(void* sendbuf, int sendcount, MPI_Datatype sendtype,
	void* recvbuf, int recvcount, MPI_Datatype recvtype, int root, MPI_Comm comm) {
	if (LinInterp_model_instance == NULL)
		return CPM_ERR_MODEL;
	return CPM_Scatter_dfs_binomial(&LinInterp_model_instance->predictor, MAX,
		sendbuf, sendcount, sendtype, recvbuf, recvcount, recvtype, root, comm);
}


/*! DFS binomial gather based on the LinInterp model. */
int LinInterp_Gather_dfs_binomial_max(void* sendbuf, int sendcount, MPI_Datatype sendtype,
	void* recvbuf, int recvcount, MPI_Datatype recvtype, int root, MPI_Comm comm) {
	if (LinInterp_model_instance == NULL)
		return CPM_ERR_MODEL;
	return CPM_Gather_dfs_binomial(&LinInterp_model_instance->predictor, MAX,
		sendbuf, sendcount, sendtype, recvbuf, recvcount, recvtype, root, comm);
}

int LinInterp_Scatter_ucs_binomial_min(void* sendbuf, int sendcount, MPI_Datatype sendtype,
	void* recvbuf, int recvcount, MPI_Datatype recvtype,
	int root, MPI_Comm comm)
{
	if (LinInterp_model_instance == NULL)
		return CPM_ERR_MODEL;
	return CPM_Scatter_ucs_binomial(&LinInterp_model_instance->predictor, MIN,
		sendbuf, sendcount, sendtype, recvbuf, recvcount, recvtype, root, comm);
}

int LinInterp_Gather_ucs_binomial_min(void* sendbuf, int sendcount, MPI_Datatype sendtype,
	void* recvbuf, int recvcount, MPI_Datatype recvtype,
	int root, MPI_Comm comm)
{
	if (LinInterp_model_instance == NULL)
		return CPM_ERR_MODEL;
	return CPM_Gather_ucs_binomial(&LinInterp_model_instance->predictor, MIN,
		sendbuf, sendcount, sendtype, recvbuf, recvcount, recvtype, root, comm);
}

int LinInterp_Scatter_ucs_binomial_max(void* sendbuf, int sendcount, MPI_Datatype sendtype,
	void* recvbuf, int recvcount, MPI_Datatype recvtype,
	int root, MPI_Comm comm)
{
	if (LinInterp_model_instance == NULL)
		return CPM_ERR_MODEL;
	return CPM_Scatter_ucs_binomial(&LinInterp_model_instance->predictor, MAX,
		sendbuf, sendcount, sendtype, recvbuf, recvcount, recvtype, root, comm);
}

int LinInterp_Gather_ucs_binomial_max(void* sendbuf, int sendcount, MPI_Datatype sendtype,
	void* recvbuf, int recvcount, MPI_Datatype recvtype,
	int root, MPI_Comm comm)
{
	if (LinInterp_model_instance == NULL)
		return CPM_ERR_MODEL;
	return CPM_Gather_ucs_binomial(&LinInterp_model_instance->predictor, MAX,
		sendbuf, sendcount, sendtype, recvbuf, recvcount, recvtype, root, comm);
}

int LinInterp_Scatter_bfs_binomial_min(void* sendbuf, int sendcount, MPI_Datatype sendtype,
	void* recvbuf, int recvcount, MPI_Datatype recvtype,
	int root, MPI_Comm comm)
{
	if (LinInterp_model_instance == NULL)
		return CPM_ERR_MODEL;
	return CPM_Scatter_bfs_binomial(&LinInterp_model_instance->predictor, MIN,
		sendbuf, sendcount, sendtype, recvbuf, recvcount, recvtype, root, comm);
}

int LinInterp_Gather_bfs_binomial_min(void* sendbuf, int sendcount, MPI_Datatype sendtype,
	void* recvbuf, int recvcount, MPI_Datatype recvtype,
	int root, MPI_Comm comm)
{
	if (LinInterp_model_instance == NULL)
		return CPM_ERR_MODEL;
	return CPM_Gather_bfs_binomial(&LinInterp_model_instance->predictor, MIN,
		sendbuf, sendcount, sendtype, recvbuf, recvcount, recvtype, root, comm);
}

int LinInterp_Scatter_bfs_binomial_max(void* sendbuf, int sendcount, MPI_Datatype sendtype,
	void* recvbuf, int recvcount, MPI_Datatype recvtype,
	int root, MPI_Comm comm)
{
	if (LinInterp_model_instance == NULL)
		return CPM_ERR_MODEL;
	return CPM_Scatter_bfs_binomial(&LinInterp_model_instance->predictor, MAX,
		sendbuf, sendcount, sendtype, recvbuf, recvcount, recvtype, root, comm);
}

int LinInterp_Gather_bfs_binomial_max(void* sendbuf, int sendcount, MPI_Datatype sendtype,
	void* recvbuf, int recvcount, MPI_Datatype recvtype,
	int root, MPI_Comm comm)
{
	if (LinInterp_model_instance == NULL)
		return CPM_ERR_MODEL;
	return CPM_Gather_bfs_binomial(&LinInterp_model_instance->predictor, MAX,
		sendbuf, sendcount, sendtype, recvbuf, recvcount, recvtype, root, comm);
}

int LinInterp_Scatterv_dfs_binomial_min(void* sendbuf, int* sendcounts, int* displs, MPI_Datatype sendtype,
	void* recvbuf, int recvcount, MPI_Datatype recvtype,
	int root, MPI_Comm comm)
{
	if (LinInterp_model_instance == NULL)
		return CPM_ERR_MODEL;
	return CPM_Scatterv_dfs_binomial(&LinInterp_model_instance->predictor, MIN,
		sendbuf, sendcounts, displs, sendtype, recvbuf, recvcount, recvtype, root, comm);
}

int LinInterp_Gatherv_dfs_binomial_min(void* sendbuf, int sendcount, MPI_Datatype sendtype,
	void* recvbuf, int* recvcounts, int* displs, MPI_Datatype recvtype,
	int root, MPI_Comm comm)
{
	if (LinInterp_model_instance == NULL)
		return CPM_ERR_MODEL;
	return CPM_Gatherv_dfs_binomial(&LinInterp_model_instance->predictor, MIN,
		sendbuf, sendcount, sendtype, recvbuf, recvcounts, displs, recvtype, root, comm);
}

int LinInterp_Scatterv_dfs_binomial_max(void* sendbuf, int* sendcounts, int* displs, MPI_Datatype sendtype,
	void* recvbuf, int recvcount, MPI_Datatype recvtype,
	int root, MPI_Comm comm)
{
	if (LinInterp_model_instance == NULL)
		return CPM_ERR_MODEL;
	return CPM_Scatterv_dfs_binomial(&LinInterp_model_instance->predictor, MAX,
		sendbuf, sendcounts, displs, sendtype, recvbuf, recvcount, recvtype, root, comm);
}

int LinInterp_Gatherv_dfs_binomial_max(void* sendbuf, int sendcount, MPI_Datatype sendtype,
	void* recvbuf, int* recvcounts, int* displs, MPI_Datatype recvtype,
	int root, MPI_Comm comm)
{
	if (LinInterp_model_instance == NULL)
		return CPM_ERR_MODEL;
	return CPM_Gatherv_dfs_binomial(&LinInterp_model_instance->predictor, MAX,
		sendbuf, sendcount, sendtype, recvbuf, recvcounts, displs, recvtype, root, comm);
}

int LinInterp_Scatterv_bfs_binomial_min(void* sendbuf, int* sendcounts, int* displs, MPI_Datatype sendtype,
	void* recvbuf, int recvcount, MPI_Datatype recvtype,
	int root, MPI_Comm comm)
{
	if (LinInterp_model_instance == NULL)
		return CPM_ERR_MODEL;
	return CPM_Scatterv_bfs_binomial(&LinInterp_model_instance->predictor, MIN,
		sendbuf, sendcounts, displs, sendtype, recvbuf, recvcount, recvtype, root, comm);
}

int LinInterp_Gatherv_bfs_binomial_min(void* sendbuf, int sendcount, MPI_Datatype sendtype,
	void* recvbuf, int* recvcounts, int* displs, MPI_Datatype recvtype,
	int root, MPI_Comm comm)
{
	if (LinInterp_model_instance == NULL)
		return CPM_ERR_MODEL;
	return CPM_Gatherv_bfs_binomial(&LinInterp_model_instance->predictor, MIN,
		sendbuf, sendcount, sendtype, recvbuf, recvcounts, displs, recvtype, root, comm);
}

int LinInterp_Scatterv_bfs_binomial_max(void* sendbuf, int* sendcounts, int* displs, MPI_Datatype sendtype,
	void* recvbuf, int recvcount, MPI_Datatype recvtype,
	int root, MPI_Comm comm)
{
	if (LinInterp_model_instance == NULL)
		return CPM_ERR_MODEL;
	return CPM_Scatterv_bfs_binomial(&LinInterp_model_instance->predictor, MAX,
		sendbuf, sendcounts, displs, sendtype, recvbuf, recvcount, recvtype, root, comm);
}

int LinInterp_Gatherv_bfs_binomial_max(void* sendbuf, int sendcount, MPI_Datatype sendtype,
	void* recvbuf, int* recvcounts, int* displs, MPI_Datatype recvtype,
	int root, MPI_Comm comm)
{
	if (LinInterp_model_instance == NULL)
		return CPM_ERR_MODEL;
	return CPM_Gatherv_bfs_binomial(&LinInterp_model_instance->predictor, MAX,
		sendbuf, sendcount, sendtype, recvbuf, recvcounts, displs, recvtype, root, comm);
}

int LinInterp_Scatterv_ucs_binomial_min(void* sendbuf, int* sendcounts, int* displs, MPI_Datatype sendtype,
	void* recvbuf, int recvcount, MPI_Datatype recvtype,
	int root, MPI_Comm comm)
{
	if (LinInterp_model_instance == NULL)
		return CPM_ERR_MODEL;
	return CPM_Scatterv_ucs_binomial(&LinInterp_model_instance->predictor, MIN,
		sendbuf, sendcounts, displs, sendtype, recvbuf, recvcount, recvtype, root, comm);
}

int LinInterp_Gatherv_ucs_binomial_min(void* sendbuf, int sendcount, MPI_Datatype sendtype,
	void* recvbuf, int* recvcounts, int* displs, MPI_Datatype recvtype,
	int root, MPI_Comm comm)
{
	if (LinInterp_model_instance == NULL)
		return CPM_ERR_MODEL;
	return CPM_Gatherv_ucs_binomial(&LinInterp_model_instance->predictor, MIN,
		sendbuf, sendcount, sendtype, recvbuf, recvcounts, displs, recvtype, root, comm);
}

int LinInterp_Scatterv_ucs_binomial_max(void* sendbuf, int* sendcounts, int* displs, MPI_Datatype sendtype,
	void* recvbuf, int recvcount, MPI_Datatype recvtype,
	int root, MPI_Comm comm)
{
	if (LinInterp_model_instance == NULL)
		return CPM_ERR_MODEL;
	return CPM_Scatterv_ucs_binomial(&LinInterp_model_instance->predictor, MAX,
		sendbuf, sendcounts, displs, sendtype, recvbuf, recvcount, recvtype, root, comm);
}

int LinInterp_Gatherv_ucs_binomial_max(void* sendbuf, int sendcount, MPI_Datatype sendtype,
	void* recvbuf, int* recvcounts, int* displs, MPI_Datatype recvtype,
	int root, MPI_Comm comm)
{
	if (LinInterp_model_instance == NULL)
		return CPM_ERR_MODEL;
	return CPM_Gatherv_ucs_binomial(&LinInterp_model_instance->predictor, MAX,
		sendbuf, sendcount, sendtype, recvbuf, recvcounts, displs, recvtype, root, comm);
}

int LinInterp_Scatterv_Traff(void* sendbuf, int* sendcounts, int* displs, MPI_Datatype sendtype,
	void* recvbuf, int recvcount, MPI_Datatype recvtype,
	int root, MPI_Comm comm)
{
	if (LinInterp_model_instance == NULL)
		return CPM_ERR_MODEL;
	return CPM_Scatterv_Traff(&LinInterp_model_instance->predictor,
		sendbuf, sendcounts, displs, sendtype,
		recvbuf, recvcount, recvtype,
		root, comm);
}

int LinInterp_Gatherv_Traff(void* sendbuf, int sendcount, MPI_Datatype sendtype,
	void* recvbuf, int* recvcounts, int* displs, MPI_Datatype recvtype,
	int root, MPI_Comm comm)
{
	if (LinInterp_model_instance == NULL)
		return CPM_ERR_MODEL;
	return CPM_Gatherv_Traff(&LinInterp_model_instance->predictor,
		sendbuf, sendcount, sendtype,
		recvbuf, recvcounts, displs, recvtype,
		root, comm);
}
