#include "config.h"
#include "lmo_collectives.h"
#include "cpm_collectives.h"
#include <memory.h>
#include "MPIBlib/collectives/mpib_basic_collectives.h"

/*!
 * \brief The global instance of the LMO model for use in the optimized scatter/gather.
 * \details Must be initialized by \ref LMO_initialize and destroyed by \ref LMO_finalize on all processes.
 */
LMO_model* LMO_model_instance = NULL;

int LMO_initialize(MPI_Comm comm, LMO_model* model)
{
	int rank;
	MPI_Comm_rank(comm, &rank);
	int n = 0;
	if (rank == 0)
		n = model->n;
	MPI_Bcast(&n, 1, MPI_INT, 0, comm); 
	LMO_model_instance = LMO_alloc(n);
	if (rank == 0)
		memcpy(LMO_model_instance->C, model->C, sizeof(double) * n);
	MPI_Bcast(LMO_model_instance->C, n, MPI_DOUBLE, 0, comm);
	if (rank == 0)
		memcpy(LMO_model_instance->t, model->t, sizeof(double) * n);
	MPI_Bcast(LMO_model_instance->t, n, MPI_DOUBLE, 0, comm);
	int Cn2 = n * (n - 1) / 2; 
	if (rank == 0)
		memcpy(LMO_model_instance->b, model->b, sizeof(double) * Cn2);
	MPI_Bcast(LMO_model_instance->b, Cn2, MPI_DOUBLE, 0, comm);
	if (rank == 0)
		LMO_model_instance->S = model->S; 
	MPI_Bcast(&LMO_model_instance->S, 1, MPI_DOUBLE, 0, comm);
	if (rank == 0)
		LMO_model_instance->M1 = model->M1; 
	MPI_Bcast(&LMO_model_instance->M1, 1, MPI_DOUBLE, 0, comm);
	if (rank == 0)
		LMO_model_instance->M2 = model->M2; 
	MPI_Bcast(&LMO_model_instance->M2, 1, MPI_DOUBLE, 0, comm);
	return 0;
}

int LMO_finalize(MPI_Comm comm)
{
	LMO_free(LMO_model_instance);
	LMO_model_instance = NULL;
	return 0;
}

int LMO_Scatter_split_flat(void* sendbuf, int sendcount, MPI_Datatype sendtype,
	void* recvbuf, int recvcount, MPI_Datatype recvtype, 
	int root, MPI_Comm comm)
{
	if (LMO_model_instance == NULL)
		return CPM_ERR_MODEL;
	int type_size;
	MPI_Type_size(sendtype, &type_size);
	int S = LMO_model_instance->S;
	int sz = sendcount * type_size; 
	if (sz > S)
	{
		int rank;
		MPI_Comm_rank(comm, &rank);
		int size;
		MPI_Comm_size(comm, &size);

		int m = 1;
		while (sz / m >= S || sz / (m - 1) < S) 
			m++;
		int count = sz / m;
		char* tmp = rank == root ? (char*)malloc(sizeof(char) * size * count) : NULL;

		int pos = 0;
		while (pos < sz)
		{
			int cnt = sz - pos < count ? sz - pos : count;
			if (rank == root)
			{
				int i;
				for (i = 0; i < size; i++)
					memcpy(tmp + cnt * i, (char*)sendbuf + sz * i + pos, cnt);
			}
			int err = MPIB_Scatter_flat_nb(tmp, cnt, MPI_CHAR, (char*)recvbuf + pos, cnt, MPI_CHAR, root, comm);
			if (err != MPI_SUCCESS)
				return err;
			pos += cnt;
		}

		if (tmp != NULL)
			free(tmp);

		return MPI_SUCCESS;
	}
	else
		return MPIB_Scatter_flat_nb(sendbuf, sendcount, sendtype, recvbuf, recvcount, recvtype, root, comm);
}

int LMO_Gather_split_flat(void* sendbuf, int sendcount, MPI_Datatype sendtype,
	void* recvbuf, int recvcount, MPI_Datatype recvtype, 
	int root, MPI_Comm comm)
{
	if (LMO_model_instance == NULL)
		return CPM_ERR_MODEL;
	int type_size;
	MPI_Type_size(sendtype, &type_size);
	int M1 = LMO_model_instance->M1;
	int M2 = LMO_model_instance->M2;
	int sz = sendcount * type_size; 
	if (sz > M1 && sz < M2)
	{
		int rank;
		MPI_Comm_rank(comm, &rank);
		int size;
		MPI_Comm_size(comm, &size);

		int m = 1;
		while (sz / m >= M1 || sz / (m - 1) < M1) 
			m++;
		int count = sz / m;
		char* tmp = rank == root ? (char*)malloc(sizeof(char) * size * count) : NULL;

		int pos = 0;
		while (pos < sz)
		{
			int cnt = sz - pos < count ? sz - pos : count;
			MPI_Barrier(comm);
			int err = MPIB_Gather_flat_nb((char*)sendbuf + pos, cnt, MPI_CHAR, tmp, cnt, MPI_CHAR, root, comm);
			if (err != MPI_SUCCESS)
				return err;
			if (rank == root)
			{
				int i;
				for (i = 0; i < size; i++)
					memcpy((char*)recvbuf + sz * i + pos, tmp + cnt * i, cnt);
			}
			pos += cnt;
		}

		if (tmp != NULL)
			free(tmp);

		return MPI_SUCCESS;
	}
	else
		return MPIB_Gather_flat_nb(sendbuf, sendcount, sendtype, recvbuf, recvcount, recvtype, root, comm);
}
