#include "config.h"
#include "plogp_collectives.h"
#include "cpm_collectives.h"
#include "logp_mpi-1.4/logp_mpi.h"
#include <string.h>

/*!
 * The global instance of PLogP model for use in the optimized scatter/gather.
 * Must be initialized by \ref PLogP_initialize and destroyed by \ref PLogP_finalize
 * at all processes.
 */
PLogP_model* PLogP_model_instance = NULL;

int PLogP_initialize(MPI_Comm comm, PLogP_model* model) {
	int rank;
	MPI_Comm_rank(comm, &rank);
	int n = rank == 0 ? model->n : 0;
	MPI_Bcast(&n, 1, MPI_INT, 0, comm); 
	PLogP_model_instance = PLogP_alloc(n);
	logp_params** logp = (logp_params**)PLogP_model_instance->logp;
	int Cn2 = n * (n - 1) / 2;
	// bcast sizes
	int i;
	for (i = 0; i< Cn2; i++) {
		logp[i] = (logp_params*)malloc(sizeof(logp_params));
		if (rank == 0)
			memcpy(logp[i], model->logp[i], sizeof(logp_params));
		logp[i]->descr = (char*)malloc(sizeof(char) * 4);
		strcpy(logp[i]->descr, "CPM");
	}
	int* blocklens = (int*)malloc(sizeof(int) * Cn2);
	MPI_Aint* displs = (MPI_Aint*)malloc(sizeof(MPI_Aint) * Cn2);
	for (i = 0; i< Cn2; i++) {
		blocklens[i] = sizeof(int) * 2 + sizeof(double);
		displs[i] = (void*)&logp[i]->nsizes - (void*)logp;
	}
	MPI_Datatype logp_type;
	MPI_Type_hindexed(Cn2, blocklens, displs, MPI_CHAR, &logp_type) ;
	MPI_Type_commit(&logp_type);
	MPI_Bcast(logp, 1, logp_type, 0, comm);
	MPI_Type_free(&logp_type);
	// bcast functions
	for (i = 0; i< Cn2; i++) {
		logp[i]->params = (logp_size_params*)malloc(sizeof(logp_size_params) * logp[i]->nsizes);
		if (rank == 0)
			memcpy(logp[i]->params, ((logp_params*)model->logp[i])->params,  sizeof(logp_size_params) * logp[i]->nsizes);
	}
	for (i = 0; i< Cn2; i++) {
		blocklens[i] = sizeof(logp_size_params) * logp[i]->nsizes;
		displs[i] = (void*)logp[i]->params - (void*)logp;
	}
	MPI_Type_hindexed(Cn2, blocklens, displs, MPI_CHAR, &logp_type) ;
	MPI_Type_commit(&logp_type);
	MPI_Bcast(logp,1, logp_type, 0, comm);
	MPI_Type_free(&logp_type);
	free(blocklens);
	free(displs);
	return 0;
}

int PLogP_finalize(MPI_Comm comm) {
	PLogP_free(PLogP_model_instance);
	PLogP_model_instance = NULL;
	return 0;
}

int PLogP_Scatterv_sorted_flat_asc(void* sendbuf, int* sendcounts, int* displs, MPI_Datatype sendtype,
	void* recvbuf, int recvcount, MPI_Datatype recvtype,
	int root, MPI_Comm comm)
{
	if (PLogP_model_instance == NULL)
		return CPM_ERR_MODEL;
	return CPM_Scatterv_sorted_flat(&PLogP_model_instance->predictor, ASC,
		sendbuf, sendcounts, displs, sendtype, recvbuf, recvcount, recvtype, root, comm);
}

int PLogP_Gatherv_sorted_flat_asc(void* sendbuf, int sendcount, MPI_Datatype sendtype,
	void* recvbuf, int* recvcounts, int* displs, MPI_Datatype recvtype,
	int root, MPI_Comm comm)
{
	if (PLogP_model_instance == NULL)
		return CPM_ERR_MODEL;
	return CPM_Gatherv_sorted_flat(&PLogP_model_instance->predictor, ASC,
		sendbuf, sendcount, sendtype, recvbuf, recvcounts, displs, recvtype, root, comm);
}

int PLogP_Scatterv_sorted_flat_dsc(void* sendbuf, int* sendcounts, int* displs, MPI_Datatype sendtype,
	void* recvbuf, int recvcount, MPI_Datatype recvtype,
	int root, MPI_Comm comm)
{
	if (PLogP_model_instance == NULL)
		return CPM_ERR_MODEL;
	return CPM_Scatterv_sorted_flat(&PLogP_model_instance->predictor, DESC,
		sendbuf, sendcounts, displs, sendtype, recvbuf, recvcount, recvtype, root, comm);
}

int PLogP_Gatherv_sorted_flat_dsc(void* sendbuf, int sendcount, MPI_Datatype sendtype,
	void* recvbuf, int* recvcounts, int* displs, MPI_Datatype recvtype,
	int root, MPI_Comm comm)
{
	if (PLogP_model_instance == NULL)
		return CPM_ERR_MODEL;
	return CPM_Gatherv_sorted_flat(&PLogP_model_instance->predictor, DESC,
		sendbuf, sendcount, sendtype, recvbuf, recvcounts, displs, recvtype, root, comm);
}

int PLogP_Bcast_dfs_binomial_min(void* buffer, int count, MPI_Datatype datatype, int root, MPI_Comm comm) {
	if (PLogP_model_instance == NULL)
		return CPM_ERR_MODEL;
	return CPM_Bcast_dfs_binomial(&PLogP_model_instance->predictor, MIN,
		buffer, count, datatype, root, comm);
}

int PLogP_Bcast_dfs_binomial_max(void* buffer, int count, MPI_Datatype datatype, int root, MPI_Comm comm) {
	if (PLogP_model_instance == NULL)
		return CPM_ERR_MODEL;
	return CPM_Bcast_dfs_binomial(&PLogP_model_instance->predictor, MAX,
		buffer, count, datatype, root, comm);
}

int PLogP_Bcast_bfs_binomial_min(void* buffer, int count, MPI_Datatype datatype, int root, MPI_Comm comm) {
	if (PLogP_model_instance == NULL)
		return CPM_ERR_MODEL;
	return CPM_Bcast_bfs_binomial(&PLogP_model_instance->predictor, MIN,
		buffer, count, datatype, root, comm);
}

int PLogP_Bcast_bfs_binomial_max(void* buffer, int count, MPI_Datatype datatype, int root, MPI_Comm comm) {
	if (PLogP_model_instance == NULL)
		return CPM_ERR_MODEL;
	return CPM_Bcast_bfs_binomial(&PLogP_model_instance->predictor, MAX,
		buffer, count, datatype, root, comm);
}

int PLogP_Reduce_dfs_binomial_min(void* sendbuf, void* recvbuf, int count, MPI_Datatype datatype,
	MPI_Op op, int root, MPI_Comm comm)
{
	if (PLogP_model_instance == NULL)
		return CPM_ERR_MODEL;
	return CPM_Reduce_dfs_binomial(&PLogP_model_instance->predictor, MIN,
		sendbuf, recvbuf, count, datatype,
		op, root, comm);
}

int PLogP_Reduce_bfs_binomial_min(void* sendbuf, void* recvbuf, int count, MPI_Datatype datatype,
	MPI_Op op, int root, MPI_Comm comm) {
	if (PLogP_model_instance == NULL)
		return CPM_ERR_MODEL;
	return CPM_Reduce_bfs_binomial(&PLogP_model_instance->predictor, MIN,
		sendbuf, recvbuf, count, datatype,
		op, root, comm);
}

int PLogP_Reduce_bfs_binomial_max(void* sendbuf, void* recvbuf, int count, MPI_Datatype datatype,
	MPI_Op op, int root, MPI_Comm comm) {
	if (PLogP_model_instance == NULL)
		return CPM_ERR_MODEL;
	return CPM_Reduce_bfs_binomial(&PLogP_model_instance->predictor, MAX,
		sendbuf, recvbuf, count, datatype,
		op, root, comm);
}

int PLogP_Reduce_dfs_binomial_max(void* sendbuf, void* recvbuf, int count, MPI_Datatype datatype,
	MPI_Op op, int root, MPI_Comm comm)
{
	if (PLogP_model_instance == NULL)
		return CPM_ERR_MODEL;
	return CPM_Reduce_dfs_binomial(&PLogP_model_instance->predictor, MAX,
		sendbuf, recvbuf, count, datatype,
		op, root, comm);
}

int PLogP_Bcast_ucs_binomial_min(void* buffer, int count, MPI_Datatype datatype, int root, MPI_Comm comm) {
	if (PLogP_model_instance == NULL)
		return CPM_ERR_MODEL;
	return CPM_Bcast_ucs_binomial(&PLogP_model_instance->predictor, MIN,
		buffer, count, datatype, root, comm);
}

int PLogP_Reduce_ucs_binomial_min(void* sendbuf, void* recvbuf, int count, MPI_Datatype datatype,
	MPI_Op op, int root, MPI_Comm comm)
{
	if (PLogP_model_instance == NULL)
		return CPM_ERR_MODEL;
	return CPM_Reduce_ucs_binomial(&PLogP_model_instance->predictor, MIN,
		sendbuf, recvbuf, count, datatype,
		op, root, comm);
}

int PLogP_Bcast_ucs_binomial_max(void* buffer, int count, MPI_Datatype datatype, int root, MPI_Comm comm) {
	if (PLogP_model_instance == NULL)
		return CPM_ERR_MODEL;
	return CPM_Bcast_ucs_binomial(&PLogP_model_instance->predictor, MAX,
		buffer, count, datatype, root, comm);
}

int PLogP_Reduce_ucs_binomial_max(void* sendbuf, void* recvbuf, int count, MPI_Datatype datatype,
	MPI_Op op, int root, MPI_Comm comm)
{
	if (PLogP_model_instance == NULL)
		return CPM_ERR_MODEL;
	return CPM_Reduce_ucs_binomial(&PLogP_model_instance->predictor, MAX,
		sendbuf, recvbuf, count, datatype,
		op, root, comm);
}

int PLogP_Scatter_dfs_binomial_min(void* sendbuf, int sendcount, MPI_Datatype sendtype,
	void* recvbuf, int recvcount, MPI_Datatype recvtype, int root, MPI_Comm comm) {
	if (PLogP_model_instance == NULL)
		return CPM_ERR_MODEL;
	return CPM_Scatter_dfs_binomial(&PLogP_model_instance->predictor, MIN,
		sendbuf, sendcount, sendtype, recvbuf, recvcount, recvtype, root, comm);
}

/*! DFS binomial gather based on the PLogP model. */
int PLogP_Gather_dfs_binomial_min(void* sendbuf, int sendcount, MPI_Datatype sendtype,
	void* recvbuf, int recvcount, MPI_Datatype recvtype, int root, MPI_Comm comm) {
	if (PLogP_model_instance == NULL)
		return CPM_ERR_MODEL;
	return CPM_Gather_dfs_binomial(&PLogP_model_instance->predictor, MIN,
		sendbuf, sendcount, sendtype, recvbuf, recvcount, recvtype, root, comm);

}

/*! DFS binomial scatter based on the PLogP model. */
int PLogP_Scatter_dfs_binomial_max(void* sendbuf, int sendcount, MPI_Datatype sendtype,
	void* recvbuf, int recvcount, MPI_Datatype recvtype, int root, MPI_Comm comm) {
	if (PLogP_model_instance == NULL)
		return CPM_ERR_MODEL;
	return CPM_Scatter_dfs_binomial(&PLogP_model_instance->predictor, MAX,
		sendbuf, sendcount, sendtype, recvbuf, recvcount, recvtype, root, comm);
}


/*! DFS binomial gather based on the PLogP model. */
int PLogP_Gather_dfs_binomial_max(void* sendbuf, int sendcount, MPI_Datatype sendtype,
	void* recvbuf, int recvcount, MPI_Datatype recvtype, int root, MPI_Comm comm) {
	if (PLogP_model_instance == NULL)
		return CPM_ERR_MODEL;
	return CPM_Gather_dfs_binomial(&PLogP_model_instance->predictor, MAX,
		sendbuf, sendcount, sendtype, recvbuf, recvcount, recvtype, root, comm);
}

int PLogP_Scatter_ucs_binomial_min(void* sendbuf, int sendcount, MPI_Datatype sendtype,
	void* recvbuf, int recvcount, MPI_Datatype recvtype,
	int root, MPI_Comm comm)
{
	if (PLogP_model_instance == NULL)
		return CPM_ERR_MODEL;
	return CPM_Scatter_ucs_binomial(&PLogP_model_instance->predictor, MIN,
		sendbuf, sendcount, sendtype, recvbuf, recvcount, recvtype, root, comm);
}

int PLogP_Gather_ucs_binomial_min(void* sendbuf, int sendcount, MPI_Datatype sendtype,
	void* recvbuf, int recvcount, MPI_Datatype recvtype,
	int root, MPI_Comm comm)
{
	if (PLogP_model_instance == NULL)
		return CPM_ERR_MODEL;
	return CPM_Gather_ucs_binomial(&PLogP_model_instance->predictor, MIN,
		sendbuf, sendcount, sendtype, recvbuf, recvcount, recvtype, root, comm);
}

int PLogP_Scatter_ucs_binomial_max(void* sendbuf, int sendcount, MPI_Datatype sendtype,
	void* recvbuf, int recvcount, MPI_Datatype recvtype,
	int root, MPI_Comm comm)
{
	if (PLogP_model_instance == NULL)
		return CPM_ERR_MODEL;
	return CPM_Scatter_ucs_binomial(&PLogP_model_instance->predictor, MAX,
		sendbuf, sendcount, sendtype, recvbuf, recvcount, recvtype, root, comm);
}

int PLogP_Gather_ucs_binomial_max(void* sendbuf, int sendcount, MPI_Datatype sendtype,
	void* recvbuf, int recvcount, MPI_Datatype recvtype,
	int root, MPI_Comm comm)
{
	if (PLogP_model_instance == NULL)
		return CPM_ERR_MODEL;
	return CPM_Gather_ucs_binomial(&PLogP_model_instance->predictor, MAX,
		sendbuf, sendcount, sendtype, recvbuf, recvcount, recvtype, root, comm);
}

int PLogP_Scatter_bfs_binomial_min(void* sendbuf, int sendcount, MPI_Datatype sendtype,
	void* recvbuf, int recvcount, MPI_Datatype recvtype,
	int root, MPI_Comm comm)
{
	if (PLogP_model_instance == NULL)
		return CPM_ERR_MODEL;
	return CPM_Scatter_bfs_binomial(&PLogP_model_instance->predictor, MIN,
		sendbuf, sendcount, sendtype, recvbuf, recvcount, recvtype, root, comm);
}

int PLogP_Gather_bfs_binomial_min(void* sendbuf, int sendcount, MPI_Datatype sendtype,
	void* recvbuf, int recvcount, MPI_Datatype recvtype,
	int root, MPI_Comm comm)
{
	if (PLogP_model_instance == NULL)
		return CPM_ERR_MODEL;
	return CPM_Gather_bfs_binomial(&PLogP_model_instance->predictor, MIN,
		sendbuf, sendcount, sendtype, recvbuf, recvcount, recvtype, root, comm);
}

int PLogP_Scatter_bfs_binomial_max(void* sendbuf, int sendcount, MPI_Datatype sendtype,
	void* recvbuf, int recvcount, MPI_Datatype recvtype,
	int root, MPI_Comm comm)
{
	if (PLogP_model_instance == NULL)
		return CPM_ERR_MODEL;
	return CPM_Scatter_bfs_binomial(&PLogP_model_instance->predictor, MAX,
		sendbuf, sendcount, sendtype, recvbuf, recvcount, recvtype, root, comm);
}

int PLogP_Gather_bfs_binomial_max(void* sendbuf, int sendcount, MPI_Datatype sendtype,
	void* recvbuf, int recvcount, MPI_Datatype recvtype,
	int root, MPI_Comm comm)
{
	if (PLogP_model_instance == NULL)
		return CPM_ERR_MODEL;
	return CPM_Gather_bfs_binomial(&PLogP_model_instance->predictor, MAX,
		sendbuf, sendcount, sendtype, recvbuf, recvcount, recvtype, root, comm);
}

int PLogP_Scatterv_dfs_binomial_min(void* sendbuf, int* sendcounts, int* displs, MPI_Datatype sendtype,
	void* recvbuf, int recvcount, MPI_Datatype recvtype,
	int root, MPI_Comm comm)
{
	if (PLogP_model_instance == NULL)
		return CPM_ERR_MODEL;
	return CPM_Scatterv_dfs_binomial(&PLogP_model_instance->predictor, MIN,
		sendbuf, sendcounts, displs, sendtype, recvbuf, recvcount, recvtype, root, comm);
}

int PLogP_Gatherv_dfs_binomial_min(void* sendbuf, int sendcount, MPI_Datatype sendtype,
	void* recvbuf, int* recvcounts, int* displs, MPI_Datatype recvtype,
	int root, MPI_Comm comm)
{
	if (PLogP_model_instance == NULL)
		return CPM_ERR_MODEL;
	return CPM_Gatherv_dfs_binomial(&PLogP_model_instance->predictor, MIN,
		sendbuf, sendcount, sendtype, recvbuf, recvcounts, displs, recvtype, root, comm);
}

int PLogP_Scatterv_dfs_binomial_max(void* sendbuf, int* sendcounts, int* displs, MPI_Datatype sendtype,
	void* recvbuf, int recvcount, MPI_Datatype recvtype,
	int root, MPI_Comm comm)
{
	if (PLogP_model_instance == NULL)
		return CPM_ERR_MODEL;
	return CPM_Scatterv_dfs_binomial(&PLogP_model_instance->predictor, MAX,
		sendbuf, sendcounts, displs, sendtype, recvbuf, recvcount, recvtype, root, comm);
}

int PLogP_Gatherv_dfs_binomial_max(void* sendbuf, int sendcount, MPI_Datatype sendtype,
	void* recvbuf, int* recvcounts, int* displs, MPI_Datatype recvtype,
	int root, MPI_Comm comm)
{
	if (PLogP_model_instance == NULL)
		return CPM_ERR_MODEL;
	return CPM_Gatherv_dfs_binomial(&PLogP_model_instance->predictor, MAX,
		sendbuf, sendcount, sendtype, recvbuf, recvcounts, displs, recvtype, root, comm);
}

int PLogP_Scatterv_bfs_binomial_min(void* sendbuf, int* sendcounts, int* displs, MPI_Datatype sendtype,
	void* recvbuf, int recvcount, MPI_Datatype recvtype,
	int root, MPI_Comm comm)
{
	if (PLogP_model_instance == NULL)
		return CPM_ERR_MODEL;
	return CPM_Scatterv_bfs_binomial(&PLogP_model_instance->predictor, MIN,
		sendbuf, sendcounts, displs, sendtype, recvbuf, recvcount, recvtype, root, comm);
}

int PLogP_Gatherv_bfs_binomial_min(void* sendbuf, int sendcount, MPI_Datatype sendtype,
	void* recvbuf, int* recvcounts, int* displs, MPI_Datatype recvtype,
	int root, MPI_Comm comm)
{
	if (PLogP_model_instance == NULL)
		return CPM_ERR_MODEL;
	return CPM_Gatherv_bfs_binomial(&PLogP_model_instance->predictor, MIN,
		sendbuf, sendcount, sendtype, recvbuf, recvcounts, displs, recvtype, root, comm);
}

int PLogP_Scatterv_bfs_binomial_max(void* sendbuf, int* sendcounts, int* displs, MPI_Datatype sendtype,
	void* recvbuf, int recvcount, MPI_Datatype recvtype,
	int root, MPI_Comm comm)
{
	if (PLogP_model_instance == NULL)
		return CPM_ERR_MODEL;
	return CPM_Scatterv_bfs_binomial(&PLogP_model_instance->predictor, MAX,
		sendbuf, sendcounts, displs, sendtype, recvbuf, recvcount, recvtype, root, comm);
}

int PLogP_Gatherv_bfs_binomial_max(void* sendbuf, int sendcount, MPI_Datatype sendtype,
	void* recvbuf, int* recvcounts, int* displs, MPI_Datatype recvtype,
	int root, MPI_Comm comm)
{
	if (PLogP_model_instance == NULL)
		return CPM_ERR_MODEL;
	return CPM_Gatherv_bfs_binomial(&PLogP_model_instance->predictor, MAX,
		sendbuf, sendcount, sendtype, recvbuf, recvcounts, displs, recvtype, root, comm);
}

int PLogP_Scatterv_ucs_binomial_min(void* sendbuf, int* sendcounts, int* displs, MPI_Datatype sendtype,
	void* recvbuf, int recvcount, MPI_Datatype recvtype,
	int root, MPI_Comm comm)
{
	if (PLogP_model_instance == NULL)
		return CPM_ERR_MODEL;
	return CPM_Scatterv_ucs_binomial(&PLogP_model_instance->predictor, MIN,
		sendbuf, sendcounts, displs, sendtype, recvbuf, recvcount, recvtype, root, comm);
}

int PLogP_Gatherv_ucs_binomial_min(void* sendbuf, int sendcount, MPI_Datatype sendtype,
	void* recvbuf, int* recvcounts, int* displs, MPI_Datatype recvtype,
	int root, MPI_Comm comm)
{
	if (PLogP_model_instance == NULL)
		return CPM_ERR_MODEL;
	return CPM_Gatherv_ucs_binomial(&PLogP_model_instance->predictor, MIN,
		sendbuf, sendcount, sendtype, recvbuf, recvcounts, displs, recvtype, root, comm);
}

int PLogP_Scatterv_ucs_binomial_max(void* sendbuf, int* sendcounts, int* displs, MPI_Datatype sendtype,
	void* recvbuf, int recvcount, MPI_Datatype recvtype,
	int root, MPI_Comm comm)
{
	if (PLogP_model_instance == NULL)
		return CPM_ERR_MODEL;
	return CPM_Scatterv_ucs_binomial(&PLogP_model_instance->predictor, MAX,
		sendbuf, sendcounts, displs, sendtype, recvbuf, recvcount, recvtype, root, comm);
}

int PLogP_Gatherv_ucs_binomial_max(void* sendbuf, int sendcount, MPI_Datatype sendtype,
	void* recvbuf, int* recvcounts, int* displs, MPI_Datatype recvtype,
	int root, MPI_Comm comm)
{
	if (PLogP_model_instance == NULL)
		return CPM_ERR_MODEL;
	return CPM_Gatherv_ucs_binomial(&PLogP_model_instance->predictor, MAX,
		sendbuf, sendcount, sendtype, recvbuf, recvcounts, displs, recvtype, root, comm);
}

int PLogP_Scatterv_Traff(void* sendbuf, int* sendcounts, int* displs, MPI_Datatype sendtype,
	void* recvbuf, int recvcount, MPI_Datatype recvtype,
	int root, MPI_Comm comm)
{
	if (PLogP_model_instance == NULL)
		return CPM_ERR_MODEL;
	return CPM_Scatterv_Traff(&PLogP_model_instance->predictor,
		sendbuf, sendcounts, displs, sendtype,
		recvbuf, recvcount, recvtype,
		root, comm);
}

int PLogP_Gatherv_Traff(void* sendbuf, int sendcount, MPI_Datatype sendtype,
	void* recvbuf, int* recvcounts, int* displs, MPI_Datatype recvtype,
	int root, MPI_Comm comm)
{
	if (PLogP_model_instance == NULL)
		return CPM_ERR_MODEL;
	return CPM_Gatherv_Traff(&PLogP_model_instance->predictor,
		sendbuf, sendcount, sendtype,
		recvbuf, recvcounts, displs, recvtype,
		root, comm);
}


