#include "cpm.h"
#include <malloc.h>
#include <gsl/gsl_statistics_double.h>

/*! List of triplets */
typedef struct CPM_triplet
{
	/*! values */
	int values[3];
	/*! previous triplet */
	struct CPM_triplet* prev;
	/*! next triplet */
	struct CPM_triplet* next;
}
CPM_triplet;

/*! List of lists of triplets */
typedef struct CPM_triplets
{
	/*! items */
	CPM_triplet* list;
	/*! previous list */
	struct CPM_triplets* prev;
	/*! next list */
	struct CPM_triplets* next;
}
CPM_triplets;

/*! Builds all combinations of non-overlapped triplets of communicating nodes */
CPM_triplets* CPM_build_triplets(int n)
{
	CPM_triplet* items = NULL;
	CPM_triplet* iter = NULL;
	int i, j, k;
	for (i = 0; i < n - 2; i++)
	{
		for (j = i + 1; j < n - 1; j++)
		{
			for (k = j + 1; k < n; k++)
			{
				if (!iter)
					(iter = items = (CPM_triplet*)malloc(sizeof(CPM_triplet)))->prev = NULL;
				else
				{
					(iter->next = (CPM_triplet*)malloc(sizeof(CPM_triplet)))->prev = iter;
					iter = iter->next; 
				}
				iter->values[0] = i;
				iter->values[1] = j;
				iter->values[2] = k;
				iter->next = NULL;
			}
		}
	}

	CPM_triplets* list = NULL;
	CPM_triplets* curr_list = NULL;
	while (items)
	{
		if (curr_list == NULL)
			(curr_list = list = (CPM_triplets*)malloc(sizeof(CPM_triplets)))->prev = NULL;
		else
		{
			(curr_list->next = (CPM_triplets*)malloc(sizeof(CPM_triplets)))->prev = curr_list;
			curr_list = curr_list->next;
		}
		curr_list->list = NULL;
		curr_list->next = NULL;

		CPM_triplet* curr_item = NULL;
		CPM_triplet* iter1 = items;
		while (iter1)
		{
			int flag = 1;
			CPM_triplet* iter2 = curr_list->list;
			while (flag && iter2)
			{
				int i, j;
				for (i = 0; i < 3; i++)
					for (j = 0; j < 3; j++)
						flag &= iter1->values[i] != iter2->values[j];
				iter2 = iter2->next;
			}
			if (flag)
			{
				if (iter1 == items)
					items = items->next;
				if (iter1->prev)
					iter1->prev->next = iter1->next;
				if (iter1->next)
					iter1->next->prev = iter1->prev;
				if (curr_item == NULL)
				{
					curr_item = curr_list->list = iter1;
					curr_item->prev = NULL;
				}
				else
				{
					(curr_item->next = iter1)->prev = curr_item;
					curr_item = iter1;
				}
				iter1 = iter1->next;
				curr_item->next = NULL;
			}
			else
				iter1 = iter1->next;
		}
	}
	return list;
}

void CPM_free_triplet(CPM_triplet* triplet)
{
	if (triplet)
	{
		CPM_free_triplet(triplet->next);
		free(triplet);
	}
}

/*! Frees the triplets recursively */
void CPM_free_triplets(CPM_triplets* triplets)
{
	if (triplets)
	{
		CPM_free_triplet(triplets->list);
		CPM_free_triplets(triplets->next);
		free(triplets);
	}
}

void CPM_measure_p2pp(MPI_Comm comm, int M, int parallel, MPIB_precision precision, MPIB_result* results)
{
	int size;
	MPI_Comm_size(comm, &size);
	int rank;
	MPI_Comm_rank(comm, &rank);

	int sendcount = 0;
	MPIB_result* res = NULL;

	double* T = (double*)malloc(sizeof(double) * precision.max_reps);
	void* buffer = malloc(M);

	CPM_triplets* list = CPM_build_triplets(size);
	CPM_triplets* iter_list = list;
	while (iter_list)
	{
		MPI_Barrier(comm);
		CPM_triplet* iter_item = iter_list->list;
		while (iter_item)
		{
			if (!parallel)
				MPI_Barrier(comm);
			int flag = 0;
			int i;
			for (i = 0; i < 3; i++)
				flag |= rank == iter_item->values[i];
			if (flag)
			{
				int i;
				for (i = 0; i < 3; i++)
				{
					int source = iter_item->values[i];
					int dests[2];
					int j, index;
					for (j = 0, index = 0; j < 3; j++)
						if (j != i)
							dests[index++] = iter_item->values[j];

					int stop = 0;
					double sum = 0;
					int reps = 0;
					double ci = 0;
					while (!stop && reps < precision.max_reps)
					{
						// because of MPI_ANY_SOURCE, different tags are used
						if (rank == source)
						{
							int i;
							for (i = 0; i < 2; i++)
								MPI_Recv(NULL, 0, MPI_CHAR, dests[i], 0, comm, MPI_STATUS_IGNORE);
							T[reps] = MPI_Wtime();
							for (i = 0; i < 2; i++)
								MPI_Send(buffer, M, MPI_CHAR, dests[i], 1, comm);
							for (i = 0; i < 2; i++)
								MPI_Recv(NULL, 0, MPI_CHAR, MPI_ANY_SOURCE, 1, comm, MPI_STATUS_IGNORE);
							sum += T[reps] = MPI_Wtime() - T[reps];
							reps++;
							if (reps >= precision.min_reps && reps > 2)
							{
								stop = (ci = MPIB_ci(precision.cl, reps, T)) * reps / sum < precision.eps;
								int i; 
								for (i = 0; i < 2; i++)
									MPI_Send(&stop, 1, MPI_INT, dests[i], 0, comm);
							} 
						}
						int i;
						for (i = 0; i < 2; i++)
						{
							if (rank == dests[i])
							{
								MPI_Send(NULL, 0, MPI_CHAR, source, 0, comm);
								MPI_Recv(buffer, M, MPI_CHAR, source, 1, comm, MPI_STATUS_IGNORE);
								MPI_Send(NULL, 0, MPI_CHAR, source, 1, comm);
								reps++;
								if (reps >= precision.min_reps && reps > 2)
									MPI_Recv(&stop, 1, MPI_INT, source, 0, comm, MPI_STATUS_IGNORE);
							}
						}
					}
					if (rank == source)
					{
						res = (MPIB_result*)realloc(res, sizeof(MPIB_result) * (sendcount + 1));
						res[sendcount].T = sum / reps;
						res[sendcount].wtick = MPI_Wtick();
						res[sendcount].reps = reps;
						res[sendcount].ci = ci;
						sendcount++;
					}
				}
			}
			iter_item = iter_item->next;
		}
		iter_list = iter_list->next;
	}
	CPM_free_triplets(list);

	free(T);
	free(buffer);

	int* recvcounts = NULL;
	int* displs = NULL;
	if(rank == 0)
		recvcounts = (int*)malloc(sizeof(int) * size);
	MPI_Gather(&sendcount, 1, MPI_INT, recvcounts, 1, MPI_INT, 0, comm);
	if(rank == 0)
	{
		displs = (int*)malloc(sizeof(int) * size);
		int i;
		for (i = 0; i < size; i++)
		{
			recvcounts[i] *= sizeof(MPIB_result);
			displs[i] = i == 0 ? 0 : displs[i - 1] + recvcounts[i - 1];
		} 
	}
	MPI_Gatherv(res, sizeof(MPIB_result) * sendcount, MPI_CHAR, rank == 0 ? results : NULL, recvcounts, displs, MPI_CHAR, 0, comm);
	free(recvcounts);
	free(displs);

	free(res);
}

void CPM_measure_p2pp_p2p(MPI_Comm comm, int M, int parallel, MPIB_precision precision, MPIB_result* results_p2pp, MPIB_result* results_p2p[2])
{
	int size;
	MPI_Comm_size(comm, &size);
	int rank;
	MPI_Comm_rank(comm, &rank);

	int sendcount = 0;
	MPIB_result* res_p2pp = NULL;
	MPIB_result* res_p2p[2] = {NULL, NULL};

	double* T2 = (double*)malloc(sizeof(double) * precision.max_reps);
	double* T1[2] = {(double*)malloc(sizeof(double) * precision.max_reps), (double*)malloc(sizeof(double) * precision.max_reps)};
	void* buffer[2] = {malloc(M), malloc(M)};

	CPM_triplets* list = CPM_build_triplets(size);
	CPM_triplets* iter_list = list;
	while (iter_list)
	{
		MPI_Barrier(comm);
		CPM_triplet* iter_item = iter_list->list;
		while (iter_item)
		{
			if (!parallel)
				MPI_Barrier(comm);
			int flag = 0;
			int i;
			for (i = 0; i < 3; i++)
				flag |= rank == iter_item->values[i];
			if (flag)
			{
				int i;
				for (i = 0; i < 3; i++)
				{
					int source = iter_item->values[i];
					int dests[2];
					int j, index;
					for (j = 0, index = 0; j < 3; j++)
						if (j != i)
							dests[index++] = iter_item->values[j];

					int stop = 0;
					double sum = 0;
					int reps = 0;
					double ci = 0;
					while (!stop && reps < precision.max_reps)
					{
						// because of MPI_ANY_SOURCE, different tags are used
						if (rank == source)
						{
							int i;
							for (i = 0; i < 2; i++)
								MPI_Recv(NULL, 0, MPI_CHAR, dests[i], 0, comm, MPI_STATUS_IGNORE);
							T2[reps] = MPI_Wtime();
							for (i = 0; i < 2; i++)
								MPI_Send(buffer[i], M, MPI_CHAR, dests[i], 1, comm);
							for (i = 0; i < 2; i++)
								MPI_Recv(NULL, 0, MPI_CHAR, MPI_ANY_SOURCE, 1, comm, MPI_STATUS_IGNORE);
							sum += T2[reps] = MPI_Wtime() - T2[reps];
							for (i = 0; i < 2; i++)
							{
								MPI_Send(NULL, 0, MPI_CHAR, dests[i], 0, comm);
								MPI_Recv(NULL, 0, MPI_CHAR, dests[i], 0, comm, MPI_STATUS_IGNORE);
								T1[i][reps] = MPI_Wtime();
								MPI_Send(buffer[i], M, MPI_CHAR, dests[i], 1, comm);
								MPI_Recv(buffer[i], M, MPI_CHAR, dests[i], 1, comm, MPI_STATUS_IGNORE);
								T1[i][reps] = MPI_Wtime() - T1[i][reps];
							}
							reps++;
							if (reps >= precision.min_reps && reps > 2)
							{
								stop = (ci = MPIB_ci(precision.cl, reps, T2)) * reps / sum < precision.eps;
								int i; 
								for (i = 0; i < 2; i++)
									MPI_Send(&stop, 1, MPI_INT, dests[i], 0, comm);
							} 
						}
						int i;
						for (i = 0; i < 2; i++)
						{
							if (rank == dests[i])
							{
								MPI_Send(NULL, 0, MPI_CHAR, source, 0, comm);
								MPI_Recv(buffer[i], M, MPI_CHAR, source, 1, comm, MPI_STATUS_IGNORE);
								MPI_Send(NULL, 0, MPI_CHAR, source, 1, comm);
								MPI_Recv(NULL, 0, MPI_CHAR, source, 0, comm, MPI_STATUS_IGNORE);
								MPI_Send(NULL, 0, MPI_CHAR, source, 0, comm);
								MPI_Recv(buffer[i], M, MPI_CHAR, source, 1, comm, MPI_STATUS_IGNORE);
								MPI_Send(buffer[i], M, MPI_CHAR, source, 1, comm);
								reps++;
								if (reps >= precision.min_reps && reps > 2)
									MPI_Recv(&stop, 1, MPI_INT, source, 0, comm, MPI_STATUS_IGNORE);
							}
						}
					}
					if (rank == source)
					{
						res_p2pp = (MPIB_result*)realloc(res_p2pp, sizeof(MPIB_result) * (sendcount + 1));
						res_p2pp[sendcount].T = sum / reps;
						res_p2pp[sendcount].wtick = MPI_Wtick();
						res_p2pp[sendcount].reps = reps;
						res_p2pp[sendcount].ci = ci;
						int i;
						for (i = 0; i < 2; i++)
						{
							res_p2p[i] = (MPIB_result*)realloc(res_p2p[i], sizeof(MPIB_result) * (sendcount + 1));
							res_p2p[i][sendcount].T = gsl_stats_mean(T1[i], 1, reps);
							res_p2p[i][sendcount].wtick = MPI_Wtick();
							res_p2p[i][sendcount].reps = reps;
							if (reps == 1)
								res_p2p[i][sendcount].ci = 0;
							else
								res_p2p[i][sendcount].ci = MPIB_ci(precision.cl, reps, T1[i]);
						}
						sendcount++;
					}
				}
			}
			iter_item = iter_item->next;
		}
		iter_list = iter_list->next;
	}
	CPM_free_triplets(list);

	free(T2);
	int i;
	for (i = 0; i < 2; i++)
	{
		free(T1[i]);
		free(buffer[i]);
	}

	int* recvcounts = NULL;
	int* displs = NULL;
	if(rank == 0)
		recvcounts = (int*)malloc(sizeof(int) * size);
	MPI_Gather(&sendcount, 1, MPI_INT, recvcounts, 1, MPI_INT, 0, comm);
	if(rank == 0)
	{
		displs = (int*)malloc(sizeof(int) * size);
		int i;
		for (i = 0; i < size; i++)
		{
			recvcounts[i] *= sizeof(MPIB_result);
			displs[i] = i == 0 ? 0 : displs[i - 1] + recvcounts[i - 1];
		} 
	}
	MPI_Gatherv(res_p2pp, sizeof(MPIB_result) * sendcount, MPI_CHAR, rank == 0 ? results_p2pp : NULL, recvcounts, displs, MPI_CHAR, 0, comm);
	for (i = 0; i < 2; i++)
		MPI_Gatherv(res_p2p[i], sizeof(MPIB_result) * sendcount, MPI_CHAR, rank == 0 ? results_p2p[i] : NULL, recvcounts, displs, MPI_CHAR, 0, comm);
	free(recvcounts);
	free(displs);

	free(res_p2pp);
	for (i = 0; i < 2; i++)
		free(res_p2p[i]);
}
