#ifndef CPM_MEASUREMENT_H_
#define CPM_MEASUREMENT_H_

/*!
 * \defgroup measurement Measurement: benchmarking specific communication experiments
 * This module extends the <a href="http://hcl.ucd.ie/project/mpiblib/api">MPIBlib</a> functionality in order to measure the execution time of
 * some specific communication experiments required to estimate the parameters of communication performance models.
 * \{
 */ 
#ifdef __cplusplus
extern "C" {
#endif

/*!
 * \brief Measures the 1-to-2 execution times.
 * \details Measures the execution times of
 * - \f$ i \xlongleftrightarrow[0]{M} jk \f$,
 * - \f$ j \xlongleftrightarrow[0]{M} ik \f$,
 * - \f$ k \xlongleftrightarrow[0]{M} ij \f$
 * 
 * in the communicator, \f$ i < j < k \f$.
 * \param comm communicator, number of nodes \f$ \ge 3 \f$
 * \param M message size
 * \param parallel several non-overlapped p2pp communications at the same time if non-zero
 * \param precision measurement parameters
 * \param results array of \f$ 3 C_n^3 \f$ measurement results (significant only at root)
 */
void CPM_measure_p2pp(MPI_Comm comm, int M, int parallel, MPIB_precision precision, MPIB_result* results);

/*!
 * \brief Measures the 1-to-2 and 1-to-1 execution times.
 * \details Measures the execution times of
 * - \f$ i \xlongleftrightarrow[0]{M} jk \f$, \f$ i \xlongleftrightarrow[0]{M} j \f$, \f$ i \xlongleftrightarrow[0]{M} k \f$,
 * - \f$ j \xlongleftrightarrow[0]{M} ik \f$, \f$ j \xlongleftrightarrow[0]{M} i \f$, \f$ j \xlongleftrightarrow[0]{M} k \f$,
 * - \f$ k \xlongleftrightarrow[0]{M} ij \f$, \f$ k \xlongleftrightarrow[0]{M} i \f$, \f$ k \xlongleftrightarrow[0]{M} j \f$
 * 
 * in the communicator, \f$ i < j < k \f$.
 * \param comm communicator, number of nodes \f$ \ge 3 \f$
 * \param M message size
 * \param parallel several non-overlapped point-to-point communications at the same time if non-zero
 * \param precision measurement parameters
 * \param results_p2pp array of \f$ 3 C_n^3 \f$ measurement results (significant only at root)
 * \param results_p2p 2 arrays of \f$ 3 C_n^3 \f$ measurement results (significant only at root)
 */
void CPM_measure_p2pp_p2p(MPI_Comm comm, int M, int parallel, MPIB_precision precision, MPIB_result* results_p2pp, MPIB_result* results_p2p[2]);

#ifdef __cplusplus
}
#endif
/*!
 * \}
 */

#endif /*CPM_MEASUREMENT_H_*/
