#ifndef CPM_MODEL_H_
#define CPM_MODEL_H_

#ifdef __cplusplus
extern "C" {
#endif

/*!
 * An predictor of the communication execution time.
 * Contains a set of the predict functions:
 * - <code>double (*predict_p2p)(void* _this, int i, int j, int M)</code> is a function
 * predicting the execution time of the point-to-point communication operation,
 * where \c i and \c j are the indices of the processors, \c M is a message size.
 * - <code>double (*predict_Y)(void* _this, int M, int root, int size)</code> is a function
 * predicting the execution time of the algorithm of collective communication operation \c Y
 * (for example, \c Scatter_binomial).
 * The \c root parameter is significant for the collective operations with the root processor.
 * 
 * Each of the estimation functions includes a self pointer argument \c _this to be used
 * in the derived data structures (communication performance models).
 */
typedef struct CPM_predictor {
	/*!
	 * Predicts the execution time of the point-to-point communication
	 * \param _this a self pointer
	 * \param i an index of a processor involved in point-to-point communication
	 * \param j an index of a processor involved in point-to-point communication
	 * \param M a message size.
	 * \return predicted execution time
	 */
	double (*predict_p2p)(void* _this, int i, int j, int M);
} CPM_predictor;

/*!
 * Invalid model instance. A return error code to be used in the model-based implementations of
 * collective communication operations.
 */
#define CPM_ERR_MODEL MPI_ERR_LASTCODE + 100

#ifdef __cplusplus
}
#endif

#endif /*CPM_MODEL_H_*/
