#ifndef CPM_MODELS_H_
#define CPM_MODELS_H_

/*!
 * \defgroup models Communication performance models
 * This module provide the following models:
 * - \ref lininterp \latexonly\cite{Hockney1994}\endlatexonly
 * - \ref hockney \latexonly\cite{Hockney1994}\endlatexonly
 * - \ref plogp \latexonly\cite{Kielmann2000}\endlatexonly
 * - \ref lmo \latexonly\cite{Lastovetsky2008ASoftwareTool}\endlatexonly
 * 
 * For each model, this module provides the following interface (\c X stands for the name of the model:
 * \c Hockney, \c PLogP or \c LMO):
 * - \code
struct X_model {
	CPM_predictor predictor;
	... // parameters
}
 * \endcode
 * is a data structure containing the parameters of the model and a set of the estimation functions
 * defined by \ref CPM_predictor.
 * - \code
X_model* X_alloc(int);
void X_free(X_model*);
 * \endcode
 * are functions to allocate/free the model data.
 * - \code
void X_estimate(MPI_Comm, MPIB_msgset, MPIB_precision, int parallel, X_model**);
 * \endcode
 * is a function responsible for estimation of the model with given precision and message sizes.
 * If the \c parallel argument is set to zero, the communication experiments will be performed
 * consequently, otherwise in parallel.
 * - \code
void X_read(FILE*, X_model**);
void X_write(FILE*, const X_model*);
 * \endcode
 * are functions for input/output of the model data.
 * 
 * This interface can be used directly in parallel applications.
 * It is also a basis for optimized implementations of MPI collective communication operations
 * (see \ref cpm_coll).
 */

#include "lininterp.h"
#include "hockney.h"
#include "plogp.h"
#include "lmo.h"

#endif /*CPM_MODELS_H_*/
