#include "config.h"
#include "hockney.h"

#include <gsl/gsl_fit.h>
#include <gsl/gsl_math.h>
#include <gsl/gsl_statistics.h>

#include "MPIBlib/benchmarks/mpib.h"
#include "MPIBlib/p2p/mpib_p2p.h"

Hockney_model* Hockney_alloc(int n) {
	Hockney_model* model = (Hockney_model*)malloc(sizeof(Hockney_model));
	model->predictor.predict_p2p = Hockney_predict_p2p;
	// to be continued: model->predictor.predict_X = Hockney_predict_X;
	model->n = n;
	model->a = (double*)malloc(sizeof(double) * MPIB_C2(n));
	model->b = (double*)malloc(sizeof(double) * MPIB_C2(n));
	return model;
}

void Hockney_free(Hockney_model* model) {
	if (model) {
		free(model->a);
		free(model->b);
		free(model);
	}
}

void Hockney_read(FILE* stream, Hockney_model** _model) {
	int n;
	fscanf(stream, "#Number of nodes: %d\n", &n);
	Hockney_model* model = *_model = Hockney_alloc(n);
	fscanf(stream, "#p2p:\ta\t\tb\n");
	int i, index;
	for(i = 0, index = 0; i < model->n - 1; i++) {
		int j;
		for(j = i + 1; j < model->n; j++, index++)
			fscanf(stream, "#%d-%d:\t%le\t%le\n", &n, &n, &model->a[index], &model->b[index]);
	}
}

void Hockney_write(FILE* stream, const Hockney_model* model) {
	fprintf(stream, "#Number of nodes: %d\n", model->n);
	fprintf(stream, "#p2p:\ta\t\tb\n");
	int i, index;
	for(i = 0, index = 0; i < model->n - 1; i++) {
		int j;
		for(j = i + 1; j < model->n; j++, index++)
			fprintf(stream, "#%d-%d:\t%le\t%le\n", i, j, model->a[index], model->b[index]);
	}
}

void Hockney_estimate(MPI_Comm comm, int M, MPIB_precision precision, int parallel, Hockney_model** _model) {
	int rank;
	MPI_Comm_rank(comm, &rank);
	int size;
	MPI_Comm_size(comm, &size);
	if (size < 2) {
		if (rank == 0)
			fprintf(stderr, "Cannot compute Hockney parameters for %d processes (must be >= 2)\n", size);
		return;
	}
	Hockney_model* model = rank == 0 ? Hockney_alloc(size) : NULL;
	MPIB_result* res_0 = (MPIB_result*)malloc(sizeof(MPIB_result) * MPIB_C2(size));
	MPIB_result* res_M = (MPIB_result*)malloc(sizeof(MPIB_result) * MPIB_C2(size));
	MPIB_p2p_container* container = MPIB_roundtrip_container_alloc(MPI_Send, MPI_Recv);
	MPIB_measure_allp2p(container, comm, parallel, 0, precision, res_0);
	MPIB_measure_allp2p(container, comm, parallel, M, precision, res_M);

	MPIB_roundtrip_container_free(container);
	if (rank == 0) {
		int i, n;
		for (i = 0, n = size * (size - 1) / 2; i < n; i++) {
			//Divide by 2 since all benchmarks send and receive the same message size
			model->a[i] = res_0[i].T / 2;
			model->b[i] = (res_M[i].T - res_0[i].T) / (2 * M);
		}
		*_model = model;
	}
	free(res_0);
	free(res_M);
	if (rank == 0)
		*_model = model;
}

void Hockney_estimate_regression(MPI_Comm comm, MPIB_msgset msgset, MPIB_precision precision, int parallel, Hockney_model** _model) {
	int rank;
	MPI_Comm_rank(comm, &rank);
	int size;
	MPI_Comm_size(comm, &size);
	if (size < 2) {
		if (rank == 0)
			fprintf(stderr, "Cannot compute Hockney parameters for %d processes (must be >= 2)\n", size);
		return;
	}
	Hockney_model* model = rank == 0 ? Hockney_alloc(size) : NULL;

	// part of the model
	int sendcount =  size - rank - 1;
	double* a = (double*)malloc(sizeof(double) * sendcount);
	double* b = (double*)malloc(sizeof(double) * sendcount);

	// measurements
	MPIB_p2p_container* container = MPIB_roundtrip_container_alloc();
	MPIB_pairs* pairs = MPIB_build_pairs(size);
	MPIB_pairs* iter_pairs = pairs;
	while (iter_pairs) {
		MPI_Barrier(comm);
		MPIB_pair* iter_pair = iter_pairs->list;
		while (iter_pair) {
			if (!parallel)
				MPI_Barrier(comm);
			int i = iter_pair->values[0];
			int j = iter_pair->values[1];
			if (rank == i || rank == j) {
				int count = 0;
				double* M = NULL ;
				double* T = NULL;
				int m;
				//calculate the number of iterations for different message sizes
				count = (msgset.max_size - msgset.min_size)/msgset.stride;
				if ((msgset.max_size - msgset.min_size) % msgset.stride > 0 ) count++ ;

				//Only process i in pair (i,j) keeps track of the benchmarks
				if (rank == i) {
					M = (double *) malloc ( count * sizeof(double) );
					T = (double *) malloc ( count * sizeof(double) );
				}
				count = 0;
				for (m = msgset.min_size; m < msgset.max_size; m += msgset.stride) {
					MPIB_result result;
					MPIB_measure_p2p(container, comm, i, j, m, precision, &result);
					if (rank == i) {
						M[count] = m;
						//Divide by 2 since we have a Sendrecv container
						T[count] = result.T / 2.;
					}
					count++;
				}
				if (rank == i) {
					double cov00, cov01, cov11, sumsq;
					gsl_fit_linear(M, 1, T, 1, count, &a[j - i - 1], &b[j - i - 1], &cov00, &cov01, &cov11, &sumsq);
					free(M);
					free(T);
				}
			}
			iter_pair = iter_pair->next;
		}
		iter_pairs = iter_pairs->next;
	}
	MPIB_free_pairs(pairs);
	MPIB_roundtrip_container_free(container);

	// allgather the results
	int* recvcounts = (int*)malloc(sizeof(int) * size);
	int* displs = (int*)malloc(sizeof(int) * size);
	int i;
	for (i = 0; i < size; i++) {
		recvcounts[i] = size - i - 1;
		displs[i] = i == 0 ? 0 : displs[i - 1] + recvcounts[i - 1];
	}
	MPI_Gatherv(a, sendcount, MPI_DOUBLE, rank == 0 ? model->a : NULL, recvcounts, displs, MPI_DOUBLE, 0, comm);
	MPI_Gatherv(b, sendcount, MPI_DOUBLE, rank == 0 ? model->b : NULL, recvcounts, displs, MPI_DOUBLE, 0, comm);
	free(recvcounts);
	free(displs);
	free(a);
	free(b);

	if (rank == 0)
		*_model = model;
}

double Hockney_predict_p2p(void* _this, int i, int j, int M) {
	Hockney_model* model = (Hockney_model*)_this;
#ifdef CPM_DEBUG
	if (i == j) {
		fprintf(stderr, "predict_p2p  with i = j not allowed!\n");
		exit(-1);
	}
#endif
	int index = MPIB_IJ2INDEX(model->n, i, j); 
#ifdef CPM_DEBUG
	if (model->a[index] + model->b[index] * M < 0.) {
		fprintf(stderr, "Negative prediction value for message size %d and process link %d-%d\n", M, i, j);
		exit(-1);
	}
#endif
	return model->a[index] + model->b[index] * M;
}

double Hockney_predict_sg_flat_serial(void* _this, int root, int M) {
	Hockney_model* model = (Hockney_model*)_this;
	double res = 0;
	int i;
	for (i = 0; i < model->n; i++)
		if (i != root)
			res += model->a[MPIB_IJ2INDEX(model->n, root, i)] +
				model->b[MPIB_IJ2INDEX(model->n, root, i)] * M;
	return res;
}

double Hockney_predict_sg_flat_parallel(void* _this, int root, int M) {
	Hockney_model* model = (Hockney_model*)_this;
	double res = 0;
	int i;
	for (i = 0; i < model->n; i++)
		if (i != root)
			res = GSL_MAX_DBL(res, model->a[MPIB_IJ2INDEX(model->n, root, i)] +
				model->b[MPIB_IJ2INDEX(model->n, root, i)] * M);
	return res;
}

double Hockney_hpredict_sg_flat_serial(void* _this, int M) {
	Hockney_model* model = (Hockney_model*)_this;
	int Cn2 = MPIB_C2(model->n);
	return (gsl_stats_mean(model->a, 1, Cn2) +
		gsl_stats_mean(model->b, 1, Cn2) * M) * (model->n - 1);
}

double Hockney_hpredict_sg_flat_parallel(void* _this, int M) {
	Hockney_model* model = (Hockney_model*)_this;
	int Cn2 = MPIB_C2(model->n);
	return gsl_stats_mean(model->a, 1, Cn2) +
		gsl_stats_mean(model->b, 1, Cn2) * M;
}

double Hockney_hpredict_sg_binomial(void* _this, int M) {
	Hockney_model* model = (Hockney_model*)_this;
	int Cn2 = MPIB_C2(model->n);
	return gsl_stats_mean(model->a, 1, Cn2) * log(model->n) / M_LN2 +
		gsl_stats_mean(model->b, 1, Cn2) * M * (model->n - 1);
}
