#ifndef HOCKNEY_H_
#define HOCKNEY_H_

#include "cpm_model.h"
#include <stdio.h>
#ifdef HAVE_CONFIG_H
#include "MPIBlib/benchmarks/mpib_measurement.h"
#else
#include "mpib_measurement.h"
#endif

/*!
 * \defgroup hockney The heterogeneous Hockney model
 * This module provides building of the heterogeneous extension of the Hockney model
 * and estimation of the execution time of point-to-point and collective communication operations.
 * 
 * In contrast to the original model \latexonly\cite{Hockney1994}\endlatexonly,
 * which is based on two point-to-point parameters, estimating the point-to-point execution time as 
 * \f$ T(M) = \alpha + \beta M \f$, the heterogeneous model distinguishes the parameters
 * of each pair of processors \f$ T_{ij}(M) = \alpha_{ij} + \beta_{ij} M \f$.
 * \{
 */
#ifdef __cplusplus
extern "C" {
#endif

/*! The heterogeneous Hockney model. The point-to-point parameters for all pairs of processors. */
typedef struct Hockney_model {
	/*! Predictor */
	CPM_predictor predictor;
	/*! Number of nodes */
	int n;
	/*! Array of \f$ C_n^2 \f$ parameters: \f$ \{\alpha_{ij}\}_{i \ne j = 0}^{n-1} \f$ */
	double* a;
	/*! Array of \f$ C_n^2 \f$ parameters: \f$ \{\beta_{ij}\}_{i \ne j = 0}^{n-1} \f$ */
	double* b;
} Hockney_model;

/*!
 * Allocates memory for the Hockney model.
 * \param n number of processors
 */
Hockney_model* Hockney_alloc(int n);

/*!
 * Frees the Hockney model.
 * \param model the Hockney model
 */
void Hockney_free(Hockney_model* model);

/*! Reads the Hockney model. */
void Hockney_read(FILE* stream, Hockney_model** model);

/*! Writes the Hockney model. */
void Hockney_write(FILE* stream, const Hockney_model* model);

/*!
 * Estimates the parameters of the Hockney model in two series of roundtrips with empty and non-empty messages.
 * (accuracy depends on the precision of measurements):
 * - Measures the execution time \f$ T_{ij}(0) \f$ of
 * each \f$ i \xlongleftrightarrow[0]{0} j \f$ roundtrip in the communicator, \f$ i < j \f$,
 * to find \f$ \alpha_{ij} = T_{ij}(0) \f$.
 * To obtain more accurate results performs a series of roundtrips and
 * takes the average \f$ T_{ij}(0) \f$.
 * - Measures the execution time \f$ T_{ij}(M) \f$ of
 * each \f$ i \xlongleftrightarrow[M]{M} j \f$ roundtrip in the communicator, \f$ i < j \f$,
 * to find \f$ \beta_{ij} = \displaystyle\frac{T_{ij}(M) - \alpha_{ij}}{M} \f$.
 * To obtain more accurate results performs a series of roundtrips and
 * takes the average \f$ T_{ij}(M) \f$.
 *
 * \param comm communicator, number of nodes \f$ \ge 2 \f$
 * \param M message size
 * \param precision measurement precision
 * \param parallel several non-overlapped point-to-point communications at the same time if non-zero
 * \param model Hockney model (significant only at root)
 */
void Hockney_estimate(MPI_Comm comm, int M, MPIB_precision precision, int parallel, Hockney_model** model);

/*!
 * Estimates the parameters of the Hockney model in a series of roundtrips with different message sizes
 * (accuracy depends on the message set):
 * - Performs single communication experiments for different message sizes.
 * - Selects message sizes regularly
 * TODO: Adaptive selection, comparing the result of measurement
 * with the prediction based on the linear regression over all previous results.
 * 
 * \param comm communicator, number of nodes \f$ \ge 2 \f$
 * \param msgset message set
 * \param precision measurement precision
 * \param parallel several non-overlapped point-to-point communications at the same time if non-zero
 * \param model Hockney model (significant only at root)
 */
void Hockney_estimate_regression(MPI_Comm comm, MPIB_msgset msgset, MPIB_precision precision, int parallel, Hockney_model** model);

/*!
 * Predicts the execution time of a point-to-point communication as
 * \f$ T_{ij}(M) = \alpha_{ij} + \beta_{ij} M \f$.
 */
double Hockney_predict_p2p(void* _this, int i, int j, int M);

/*!
 * Heterogeneous prediction of flat-tree scatter/gather (sequential point-to-point communications)
 * \f$ \displaystyle\sum_{i = 0, i \neq r}^{n - 1}(\alpha_{ri} + \beta_{ri} M) \f$
 */
double Hockney_predict_sg_flat_serial(void* _this, int root, int M);

/*!
 * Heterogeneous prediction of flat-tree scatter/gather (parallel point-to-point communications)
 * \f$ \displaystyle\max_{i = 0, i \neq r}^{n - 1}(\alpha_{ri} + \beta_{ri} M) \f$
 */
double Hockney_predict_sg_flat_parallel(void* _this, int root, int M);

/*!
 * Heterogeneous prediction of binomial scatter/gather (point-to-point communications are
 * sequential within the same root or parallel otherwise)
 * \f$ \displaystyle\overbrace{\alpha_{ri} + \beta_{ri} 2^{log_2(n - 1)} M + \max S(log_2(n - 1) - 1)} \f$
 */
double Hockney_predict_sg_binomial(void* _this, int root, int M);

/*!
 * Homogeneous prediction of flat-tree scatter/gather (sequential point-to-point communications)
 * \f$ (n - 1)(\alpha + \beta M) \f$
 */
double Hockney_hpredict_sg_flat_serial(void* _this, int M);

/*!
 * Homogeneous prediction of flat-tree scatter/gather (parallel point-to-point communications)
 * \f$ \alpha + \beta M \f$
 */
double Hockney_hpredict_sg_flat_parallel(void* _this, int M);

/*!
 * Homogeneous prediction of binomial scatter/gather (point-to-point communications are
 * sequential within the same root or parallel otherwise)
 * \f$ (log_2 n) \alpha + (n - 1) \beta M \f$
 */
double Hockney_hpredict_sg_binomial(void* _this, int M);

#ifdef __cplusplus
}
#endif
/*!
 * \}
 */

#endif /*HOCKNEY_H_*/
