#include "config.h"
#include "lmo.h"
#include <gsl/gsl_math.h>
#include "cpm_output.h"
#include <gsl/gsl_fit.h>
#ifdef HAVE_LIBR
#include "cpm_R.h"
#include <Rdefines.h>
#include <R_ext/Parse.h>
#endif /*HAVE_LIBR*/
#include "cpm_measurement.h"
#include "MPIBlib/benchmarks/mpib.h"
#include "MPIBlib/collectives/mpib_basic_collectives.h"

LMO_model* LMO_alloc(int n)
{
	LMO_model* model = (LMO_model*)malloc(sizeof(LMO_model));
	model->predictor.predict_p2p = LMO_predict_p2p;
	// to be continued: model->predictor.predict_X = LMO_predict_X;
	model->n = n;
	model->C = (double*)calloc(n, sizeof(double));
	model->L = (double*)calloc(MPIB_C2(n), sizeof(double));
	model->t = (double*)calloc(n, sizeof(double));
	model->b = (double*)calloc(MPIB_C2(n), sizeof(double));
	return model;
}

void LMO_free(LMO_model* model)
{
	if (model)
	{
		free(model->C);
		free(model->L);
		free(model->t);
		free(model->b);
		free(model);
	}
}

void LMO_read(FILE* stream, LMO_model** _model)
{
	int n;
	fscanf(stream, "#Number of nodes: %d\n", &n);
	LMO_model* model = *_model = LMO_alloc(n);
	fscanf(stream, "#Fixed processing delays:");
	int i;
	for (i = 0; i < model->n; i++)
		fscanf(stream, "\t%le", &model->C[i]);
	fscanf(stream, "\n#Latencies:");
	int Cn2 = model->n * (model->n - 1) / 2;
	for (i = 0; i < Cn2; i++)
		fscanf(stream, "\t%le", &model->L[i]);
	fscanf(stream, "\n#Variable processing delays:");
	for (i = 0; i < model->n; i++)
		fscanf(stream, "\t%le", &model->t[i]);
	fscanf(stream, "\n#Transmission rates:");
	for (i = 0; i < Cn2; i++)
		fscanf(stream, "\t%le", &model->b[i]);
	fscanf
	(
		stream,
"\n#S: %d\n\
#one2many linear model for small message sizes: %le %le\n\
#one2many linear model for large message sizes: %le %le\n\
#M1: %d\n\
#M2: %d\n\
#many2one linear model for small message sizes: %le %le\n\
#many2one linear model for large message sizes: %le %le\n",
		&model->S,
		&model->one2many_small[0], &model->one2many_small[1],
		&model->one2many_large[0], &model->one2many_large[1],
		&model->M1,
		&model->M2,
		&model->many2one_small[0], &model->many2one_small[1],
		&model->many2one_large[0], &model->many2one_large[1]
	);
}

void LMO_write(FILE* stream, const LMO_model* model)
{
	fprintf(stream, "#Number of nodes: %d\n", model->n);
	fprintf(stream, "#Fixed processing delays:");
	int i;
	for (i = 0; i < model->n; i++)
		fprintf(stream, "\t%le", model->C[i]);
	int Cn2 = model->n * (model->n - 1) / 2;
	fprintf(stream, "\n#Latencies:");
	for (i = 0; i < Cn2; i++)
		fprintf(stream, "\t%le", model->L[i]);
	fprintf(stream, "\n#Variable processing delays:");
	for (i = 0; i < model->n; i++)
		fprintf(stream, "\t%le", model->t[i]);
	fprintf(stream, "\n#Transmission rates:");
	for (i = 0; i < Cn2; i++)
		fprintf(stream, "\t%le", model->b[i]);
	fprintf
	(
		stream,
"\n#S: %d\n\
#one2many linear model for small message sizes: %le %le\n\
#one2many linear model for large message sizes: %le %le\n\
#M1: %d\n\
#M2: %d\n\
#many2one linear model for small message sizes: %le %le\n\
#many2one linear model for large message sizes: %le %le\n",
		model->S,
		model->one2many_small[0], model->one2many_small[1],
		model->one2many_large[0], model->one2many_large[1],
		model->M1, model->M2,
		model->many2one_small[0], model->many2one_small[1],
		model->many2one_large[0], model->many2one_large[1]
	);
}

void LMO_estimate_p2p(LMO_model* model, MPI_Comm comm, int parallel, MPIB_precision precision)
{
	int rank;
	MPI_Comm_rank(comm, &rank);
	int size;
	MPI_Comm_size(comm, &size);
	if (size < 3)
	{
		if (rank == 0)
			fprintf(stderr, "Cannot compute p2p parameters for %d processes (must be >= 3)\n", size);
		return;
	}

	MPIB_p2p_container* container = MPIB_roundtrip_container_alloc(MPI_Send, MPI_Recv);
	int n = size;
	int M = rank == 0 ? model->S : 0;
	MPI_Bcast(&M, 1, MPI_INT, 0, comm);
	MPIB_result* res_p2p_0 = (MPIB_result*)malloc(sizeof(MPIB_result) * MPIB_C2(n));
	MPIB_measure_allp2p(container, comm, 0, parallel, precision, res_p2p_0);
#ifdef CPM_DEBUG
	if (rank == 0)
		MPIB_print_p2p_table(0, parallel, precision, n, res_p2p_0);
#endif
	MPIB_result* res_p2pp_0 = (MPIB_result*)malloc(sizeof(MPIB_result) * 3 * LMO_C3(n));
	CPM_measure_p2pp(comm, 0, parallel, precision, res_p2pp_0);
#ifdef CPM_DEBUG
	if (rank == 0)
		CPM_print_p2pp_results(n, 0, precision, res_p2pp_0, res_p2p_0, res_p2p_0);
#endif
	MPIB_result* res_p2p_M = (MPIB_result*)malloc(sizeof(MPIB_result) * MPIB_C2(n));
	MPIB_measure_allp2p(container, comm, M, parallel, precision, res_p2p_M);
#ifdef CPM_DEBUG
	if (rank == 0)
		MPIB_print_p2p_table(M, parallel, precision, n, res_p2p_M);
#endif
	MPIB_result* res_p2pp_M = (MPIB_result*)malloc(sizeof(MPIB_result) * 3 * LMO_C3(n));
	CPM_measure_p2pp(comm, M, parallel, precision, res_p2pp_M);
#ifdef CPM_DEBUG
	if (rank == 0)
		CPM_print_p2pp_results(n, M, precision, res_p2pp_M, res_p2p_0, res_p2p_M);
#endif
	MPIB_roundtrip_container_free(container);

	if (rank == 0)
	{
		memset(model->C, 0, sizeof(double) * n);
		memset(model->L, 0, sizeof(double) * MPIB_C2(n));
		memset(model->t, 0, sizeof(double) * n);
		memset(model->b, 0, sizeof(double) * MPIB_C2(n));
		int i, j, k, index;
		for (i = 0, index = 0; i < n - 2; i++)
		{
			for (j = i + 1; j < n - 1; j++)
			{
				for (k = j + 1; k < n; k++, index++)
				{
					double Ci = (res_p2pp_0[3 * index + 0].T - GSL_MAX_DBL(res_p2p_0[MPIB_IJ2INDEX(n, i, j)].T, res_p2p_0[MPIB_IJ2INDEX(n, i, k)].T)) / 2;
					double Cj = (res_p2pp_0[3 * index + 1].T - GSL_MAX_DBL(res_p2p_0[MPIB_IJ2INDEX(n, i, j)].T, res_p2p_0[MPIB_IJ2INDEX(n, j, k)].T)) / 2;
					double Ck = (res_p2pp_0[3 * index + 2].T - GSL_MAX_DBL(res_p2p_0[MPIB_IJ2INDEX(n, i, k)].T, res_p2p_0[MPIB_IJ2INDEX(n, j, k)].T)) / 2;
					double Lij = res_p2p_0[MPIB_IJ2INDEX(n, i, j)].T / 2 - Ci - Cj;
					double Ljk = res_p2p_0[MPIB_IJ2INDEX(n, j, k)].T / 2 - Cj - Ck;
					double Lik = res_p2p_0[MPIB_IJ2INDEX(n, i, k)].T / 2 - Ci - Ck;
					double ti = (res_p2pp_M[3 * index + 0].T - GSL_MAX_DBL(
						res_p2p_0[MPIB_IJ2INDEX(n, i, j)].T + res_p2p_M[MPIB_IJ2INDEX(n, i, j)].T,
						res_p2p_0[MPIB_IJ2INDEX(n, i, k)].T + res_p2p_M[MPIB_IJ2INDEX(n, i, k)].T
						) / 2 - Ci * 2) / M;
					double tj = (res_p2pp_M[3 * index + 1].T - GSL_MAX_DBL(
						res_p2p_0[MPIB_IJ2INDEX(n, i, j)].T + res_p2p_M[MPIB_IJ2INDEX(n, i, j)].T,
						res_p2p_0[MPIB_IJ2INDEX(n, j, k)].T + res_p2p_M[MPIB_IJ2INDEX(n, j, k)].T
						) / 2 - Cj * 2) / M;
					double tk = (res_p2pp_M[3 * index + 2].T - GSL_MAX_DBL(
						res_p2p_0[MPIB_IJ2INDEX(n, i, k)].T + res_p2p_M[MPIB_IJ2INDEX(n, i, k)].T,
						res_p2p_0[MPIB_IJ2INDEX(n, j, k)].T + res_p2p_M[MPIB_IJ2INDEX(n, j, k)].T
						) / 2 - Ck * 2) / M;
					model->C[i] += Ci;
					model->C[j] += Cj;
					model->C[k] += Ck;
					model->L[MPIB_IJ2INDEX(n, i, j)] += Lij; 
					model->L[MPIB_IJ2INDEX(n, j, k)] += Ljk; 
					model->L[MPIB_IJ2INDEX(n, i, k)] += Lik; 
					model->t[i] += ti;
					model->t[j] += tj;
					model->t[k] += tk;
					model->b[MPIB_IJ2INDEX(n, i, j)] += (res_p2p_M[MPIB_IJ2INDEX(n, i, j)].T / 2 - Ci - Lij - Cj) / M - ti -tj;
					model->b[MPIB_IJ2INDEX(n, j, k)] += (res_p2p_M[MPIB_IJ2INDEX(n, j, k)].T / 2 - Cj - Ljk - Ck) / M - tj -tk;
					model->b[MPIB_IJ2INDEX(n, i, k)] += (res_p2p_M[MPIB_IJ2INDEX(n, i, k)].T / 2 - Ci - Lik - Ck) / M - ti -tk;
				}
			}
		}
		for (i = 0; i < n; i++)
		{
			model->C[i] /= MPIB_C2(n - 1);
			model->t[i] /= MPIB_C2(n - 1);
		}
		for (i = 0, index = 0; i < n - 1; i++)
		{
			for (j = i + 1; j < n; j++, index++)
			{
				model->L[index] /= n - 2;
				model->b[index] /= n - 2;
				model->b[index] = 1.0 / model->b[index];
			}
		}
	}

	free(res_p2p_0);
	free(res_p2pp_0);
	free(res_p2p_M);
	free(res_p2pp_M);
}

void LMO_estimate_p2p_(LMO_model* model, MPI_Comm comm, int parallel, MPIB_precision precision)
{
	int rank;
	MPI_Comm_rank(comm, &rank);
	int size;
	MPI_Comm_size(comm, &size);
	if (size < 3)
	{
		if (rank == 0)
			fprintf(stderr, "Cannot compute p2p parameters for %d processes (must be >= 3)\n", size);
		return;
	}

	int n = size;
	int M = rank == 0 ? model->S : 0;
	MPI_Bcast(&M, 1, MPI_INT, 0, comm);

	MPIB_result* res_p2pp_0 = (MPIB_result*)malloc(sizeof(MPIB_result) * 3 * LMO_C3(n));
	MPIB_result* res_p2p_0[2] = {
		(MPIB_result*)malloc(sizeof(MPIB_result) * 3 * LMO_C3(n)),
		(MPIB_result*)malloc(sizeof(MPIB_result) * 3 * LMO_C3(n))
	};
	CPM_measure_p2pp_p2p(comm, 1, parallel, precision, res_p2pp_0, res_p2p_0);
#ifdef CPM_DEBUG
	if (rank == 0)
		CPM_print_p2pp_results_(n, 1, precision, res_p2pp_0, res_p2p_0, res_p2p_0);
#endif
	MPIB_result* res_p2pp_M = rank == 0 ? (MPIB_result*)malloc(sizeof(MPIB_result) * 3 * LMO_C3(n)) : NULL;
	MPIB_result* res_p2p_M[2] = {
		rank == 0 ? (MPIB_result*)malloc(sizeof(MPIB_result) * 3 * LMO_C3(n)) : NULL,
		rank == 0 ? (MPIB_result*)malloc(sizeof(MPIB_result) * 3 * LMO_C3(n)) : NULL
	};
	CPM_measure_p2pp_p2p(comm, M, parallel, precision, res_p2pp_M, res_p2p_M);
#ifdef CPM_DEBUG
	if (rank == 0)
		CPM_print_p2pp_results_(n, M, precision, res_p2pp_M, res_p2p_0, res_p2p_M);
#endif

	if (rank == 0)
	{
		memset(model->C, 0, sizeof(double) * n);
		memset(model->L, 0, sizeof(double) * MPIB_C2(n));
		memset(model->t, 0, sizeof(double) * n);
		memset(model->b, 0, sizeof(double) * MPIB_C2(n));
		int i, j, k, index;
		for (i = 0, index = 0; i < n - 2; i++)
		{
			for (j = i + 1; j < n - 1; j++)
			{
				for (k = j + 1; k < n; k++, index++)
				{
					double Ci = (res_p2pp_0[3 * index + 0].T - GSL_MAX_DBL(res_p2p_0[0][3 * index + 0].T, res_p2p_0[1][3 * index + 0].T)) / 2;
					double Cj = (res_p2pp_0[3 * index + 1].T - GSL_MAX_DBL(res_p2p_0[0][3 * index + 1].T, res_p2p_0[0][3 * index + 1].T)) / 2;
					double Ck = (res_p2pp_0[3 * index + 2].T - GSL_MAX_DBL(res_p2p_0[0][3 * index + 2].T, res_p2p_0[1][3 * index + 2].T)) / 2;
					double Lij = (res_p2p_0[0][3 * index + 0].T + res_p2p_0[0][3 * index + 1].T) / 4 - Ci - Cj;
					double Ljk = (res_p2p_0[1][3 * index + 1].T + res_p2p_0[1][3 * index + 2].T) / 4 - Ci - Cj;
					double Lik = (res_p2p_0[1][3 * index + 0].T + res_p2p_0[0][3 * index + 2].T) / 4 - Ci - Cj;
					double ti = (res_p2pp_M[3 * index + 0].T - GSL_MAX_DBL(
						res_p2p_0[0][3 * index + 0].T + res_p2p_M[0][3 * index + 0].T,
						res_p2p_0[1][3 * index + 0].T + res_p2p_M[1][3 * index + 0].T
						) / 2 - Ci * 2) / M;
					double tj = (res_p2pp_M[3 * index + 1].T - GSL_MAX_DBL(
						res_p2p_0[0][3 * index + 1].T + res_p2p_M[0][3 * index + 1].T,
						res_p2p_0[1][3 * index + 1].T + res_p2p_M[1][3 * index + 1].T
						) / 2 - Cj * 2) / M;
					double tk = (res_p2pp_M[3 * index + 2].T - GSL_MAX_DBL(
						res_p2p_0[0][3 * index + 2].T + res_p2p_M[0][3 * index + 2].T,
						res_p2p_0[1][3 * index + 2].T + res_p2p_M[1][3 * index + 2].T
						) / 2 - Ck * 2) / M;
					model->C[i] += Ci;
					model->C[j] += Cj;
					model->C[k] += Ck;
					model->L[MPIB_IJ2INDEX(n, i, j)] += Lij; 
					model->L[MPIB_IJ2INDEX(n, j, k)] += Ljk; 
					model->L[MPIB_IJ2INDEX(n, i, k)] += Lik; 
					model->t[i] += ti;
					model->t[j] += tj;
					model->t[k] += tk;
					model->b[MPIB_IJ2INDEX(n, i, j)] += ((res_p2p_M[0][3 * index + 0].T + res_p2p_M[0][3 * index + 1].T) / 4 - Ci - Lij - Cj) / M - ti -tj;
					model->b[MPIB_IJ2INDEX(n, j, k)] += ((res_p2p_M[1][3 * index + 1].T + res_p2p_M[1][3 * index + 2].T) / 4 - Cj - Ljk - Ck) / M - tj -tk;
					model->b[MPIB_IJ2INDEX(n, i, k)] += ((res_p2p_M[1][3 * index + 0].T + res_p2p_M[0][3 * index + 2].T) / 4 - Ci - Lik - Ck) / M - ti -tk;
				}
			}
		}
		for (i = 0; i < n; i++)
		{
			model->C[i] /= MPIB_C2(n - 1);
			model->t[i] /= MPIB_C2(n - 1);
		}
		for (i = 0, index = 0; i < n - 1; i++)
		{
			for (j = i + 1; j < n; j++, index++)
			{
				model->L[index] /= n - 2;
				model->b[index] /= n - 2;
				model->b[index] = 1.0 / model->b[index];
			}
		}
	}

	free(res_p2pp_0);
	free(res_p2p_0[0]);
	free(res_p2p_0[1]);
	free(res_p2pp_M);
	free(res_p2p_M[0]);
	free(res_p2p_M[1]);
}

#ifdef HAVE_LIBR
void LMO_estimate_one2many(LMO_model* model, MPI_Comm comm, int stride, int max_size, MPIB_precision precision)
{
	int rank;
	MPI_Comm_rank(comm, &rank);

	int n = max_size / stride;
	int* M = (int*)malloc(sizeof(int) * n);
	MPIB_result* results = (MPIB_result*)malloc(sizeof(MPIB_result) * n);
	MPIB_coll_container* container = (MPIB_coll_container*)MPIB_Scatter_container_alloc(MPIB_Scatter_flat_nb);
#ifdef CPM_DEBUG
	if (rank == 0)
		MPIB_print_coll_th("MPIB_Scatter_linear", "max", model->n, 0, precision);
#endif
	int i, size;
	for (i = 0, size = 0; i < n; i++, size += stride)
	{
		M[i] = size;
		MPIB_measure_max(container, comm, 0, M[i], precision, &results[i]);
#ifdef CPM_DEBUG
		if (rank == 0)
			MPIB_print_coll_tr(M[i], results[i]);
#endif
	}
	MPIB_coll_container_free(container);

	if (rank == 0)
	{
		SEXP R_M;
		PROTECT(R_M = allocVector(INTSXP, n));
		int i;
		for(i = 0; i < n; i++)
			INTEGER(R_M)[i]  = M[i];
		defineVar(install("M"), R_M, R_GlobalEnv);
		SEXP R_T;
		PROTECT(R_T = allocVector(REALSXP, n));
		for(i = 0; i < n; i++)
			REAL(R_T)[i]  = results[i].T;
		defineVar(install("T"), R_T, R_GlobalEnv);
		SEXP formula;
		ParseStatus status;
		PROTECT(formula = EVAL(VECTOR_ELT(R_ParseVector(mkString("T~M"), 1, &status, R_NilValue), 0)));
		SEXP fstats;
		PROTECT(fstats = EVAL(lang2(install("Fstats"), formula)));
		SEXP breakpoints;
		PROTECT(breakpoints = EVAL(lang2(install("breakpoints"), fstats)));
		int index = (int)REAL(CPM_getListElement(breakpoints, "breakpoints"))[0];
		UNPROTECT(5);

		model->S = M[index];
		double* x = (double*)malloc(sizeof(double) * n);
		double* y = (double*)malloc(sizeof(double) * n);
		for(i = 0; i < n; i++)
		{
			x[i] = M[i];
			y[i] = results[i].T;
		}
		double cov00, cov01, cov11, sumsq;
		gsl_fit_linear(x, 1, y, 1, index, &model->one2many_small[0], &model->one2many_small[1], &cov00, &cov01, &cov11, &sumsq);
		gsl_fit_linear(&x[index], 1, &y[index], 1, n - index, &model->one2many_large[0], &model->one2many_large[1], &cov00, &cov01, &cov11, &sumsq);
		free(x);
		free(y);
	}

	free(M);
	free(results);
}

void LMO_estimate_many2one(LMO_model* model, MPI_Comm comm, int stride, int max_size, MPIB_precision precision)
{
	int rank;
	MPI_Comm_rank(comm, &rank);

	MPIB_coll_container* container = (MPIB_coll_container*)MPIB_Gather_container_alloc(MPIB_Gather_flat_nb);

	int n = max_size / stride;
	int* M = (int*)malloc(sizeof(int) * n);
	MPIB_result* results = (MPIB_result*)malloc(sizeof(MPIB_result) * n);
#ifdef CPM_DEBUG
	if (rank == 0)
		MPIB_print_coll_th("MPIB_Gather_linear", "max", model->n, 0, precision);
#endif
	int i, size;
	for (i = 0, size = 0; i < n; i++, size += stride)
	{
		M[i] = size;
		MPIB_measure_max(container, comm, 0, M[i], precision, &results[i]);
#ifdef CPM_DEBUG
		if (rank == 0)
			MPIB_print_coll_tr(M[i], results[i]);
#endif
	}

	int M1 = 0;
	if (rank == 0)
	{
		SEXP R_T;
		PROTECT(R_T = allocVector(REALSXP, n));
		for(i = 0; i < n; i++)
			REAL(R_T)[i]  = results[i].T;
		defineVar(install("T"), R_T, R_GlobalEnv);
		SEXP formula;
		ParseStatus status;
		PROTECT(formula = EVAL(VECTOR_ELT(R_ParseVector(mkString("T~1"), 1, &status, R_NilValue), 0)));
		SEXP fstats;
		PROTECT(fstats = EVAL(lang2(install("Fstats"), formula)));
		SEXP breakpoints;
		PROTECT(breakpoints = EVAL(lang2(install("breakpoints"), fstats)));
		int index = (int)REAL(CPM_getListElement(breakpoints, "breakpoints"))[0];
		UNPROTECT(4);

		model->M2 = M[index];
		double* x = (double*)malloc(sizeof(double) * (n - index));
		double* y = (double*)malloc(sizeof(double) * (n - index));
		for(i = 0; i < n - index; i++)
		{
			x[i] = M[i + index];
			y[i] = results[i + index].T;
		}
		double cov00, cov01, cov11, sumsq;
		gsl_fit_linear(x, 1, y, 1, n - index, &model->many2one_large[0], &model->many2one_large[1], &cov00, &cov01, &cov11, &sumsq);
		free(x);
		free(y);

		for (i = 1; i < index; i++)
			if (M[i] > 1024 && results[i].T / results[i - 1].T > 10)
				break;
		M1 = M[i];
	}
	//The following section tries to find a proper M1. 
	//However, it takes very long time on some platforms and is commented for now
	/*
	MPI_Bcast(&M1, 1, MPI_INT, 0, comm);
	stride /= 2;

	while (stride > 0)
	{
		free(M);
		free(results);

		int n = (M1 - 1024) / stride;
		M = (int*)malloc(sizeof(int) * n);
		results = (MPIB_result*)malloc(sizeof(MPIB_result) * n);
		int i, size;
		for (i = 0, size = 1024; i < n; i++, size += stride)
		{
			M[i] = size;
			MPIB_measure_max(container, comm, 0, M[i], precision, &results[i]);
#ifdef CPM_DEBUG
			if (rank == 0)
				MPIB_print_coll_tr(M[i], results[i]);
#endif
		}

		if (rank == 0)
		{
			int i;
			for (i = 1; i < n; i++)
				if (results[i].T / results[i - 1].T > 10)
					break;
			if (i < n)
				M1 = M[i];
		}
		MPI_Bcast(&M1, 1, MPI_INT, 0, comm);
		stride /= 2;
	}
	*/
	MPIB_coll_container_free(container);

	if (rank == 0)
	{
		model->M1 = M1 - M1 % 1024;
		int i;
		for (i = 0; i < n; i++)
			if (M[i] >= model->M1)
				break;
		int index = i;
		double* x = (double*)malloc(sizeof(double) * index);
		double* y = (double*)malloc(sizeof(double) * index);
		for(i = 0; i < index; i++)
		{
			x[i] = M[i];
			y[i] = results[i].T;
		}
		double cov00, cov01, cov11, sumsq;
		gsl_fit_linear(x, 1, y, 1, index, &model->many2one_small[0], &model->many2one_small[1], &cov00, &cov01, &cov11, &sumsq);
		free(x);
		free(y);
	}

	free(M);
	free(results);
}
#endif /*HAVE_LIBR*/

void LMO_estimate(MPI_Comm comm, MPIB_precision precision, MPIB_msgset msgset, int parallel,
	LMO_model** model)
{
	int rank;
	MPI_Comm_rank(comm, &rank);
	int size;
	MPI_Comm_size(comm, &size);
	*model = (rank == 0 ? LMO_alloc(size) : NULL);
#ifdef HAVE_LIBR
	LMO_estimate_one2many(*model, comm, msgset.stride, msgset.max_size, precision);
	LMO_estimate_many2one(*model, comm, msgset.stride, msgset.max_size, precision);
#endif /*HAVE_LIBR*/
	double time = MPI_Wtime();
	LMO_estimate_p2p_(*model, comm, parallel, precision);
	time = MPI_Wtime() - time;
	if (rank == 0)
		printf("#LMO p2p time: %le\n", time);
	
}

void LMO_estimate_homogeneous(MPI_Comm comm, MPIB_precision precision, MPIB_msgset msgset, LMO_model** model_) {
	int rank;
	MPI_Comm_rank(comm, &rank);
	LMO_model* model = (rank == 0 ? LMO_alloc(1) : NULL);
	// todo
	*model_ = model;
}

double LMO_predict_p2p(void* _this, int i, int j, int M) {
	LMO_model* model = (LMO_model*)_this;
	int index = MPIB_IJ2INDEX(model->n, i, j);
	return model->C[i] +  model->L[index] + model->C[j] + (model->t[i] + 1.0 / model->b[index] + model->t[j]) * M; 
}

double LMO_predict_scatter_flat(void* _this, int root, int M) {
	LMO_model* model = (LMO_model*)_this;
	double res = (model->C[root] + model->t[root] * M) * (model->n - 1);
	if (M < model->S) {
		double max = 0;
		int i;
		for (i = 0; i < model->n; i++) {
			if (i != root) { 
				int index = MPIB_IJ2INDEX(model->n, root, i);
				double tmp = model->L[index] + model->C[i] + (1.0 / model->b[index] + model->t[i]) * M;
				if (tmp > max)
					max = tmp;
			}
		}
		res += max;
	}
	else {
		double max = 0;
		int i;
		for (i = 0; i < model->n; i++) {
			if (i != root) { 
				int index = MPIB_IJ2INDEX(model->n, root, i);
				double tmp = model->L[index] + model->C[i] + (1.0 / model->b[index] + model->t[i]) * model->S;
				if (tmp > max)
					max = tmp;
			}
		}
		res += max;
		for (i = 0; i < model->n; i++) {
			if (i != root) { 
				int index = MPIB_IJ2INDEX(model->n, root, i);
				res += model->L[index] + model->C[i] + (1.0 / model->b[index] + model->t[i]) * (M - model->S);
			}
		}
	}
	return res;
}

double LMO_predict_gather_flat(void* _this, int root, int M) {
	LMO_model* model = (LMO_model*)_this;
	double res = (model->C[root] + model->t[root] * M) * (model->n - 1);
	if (M < model->M1) {
		double max = 0;
		int i;
		for (i = 0; i < model->n; i++) {
			if (i != root) { 
				int index = MPIB_IJ2INDEX(model->n, root, i);
				double tmp = model->L[index] + model->C[i] + (1.0 / model->b[index] + model->t[i]) * M;
				if (tmp > max)
					max = tmp;
			}
		}
		res += max;
	}
	if (M > model->M2) {
		double max = 0;
		int i;
		for (i = 0; i < model->n; i++) {
			if (i != root) { 
				int index = MPIB_IJ2INDEX(model->n, root, i);
				double tmp = model->L[index] + model->C[i] + (1.0 / model->b[index] + model->t[i]) * model->M2;
				if (tmp > max)
					max = tmp;
			}
		}
		res += max;
		for (i = 0; i < model->n; i++) {
			if (i != root) { 
				int index = MPIB_IJ2INDEX(model->n, root, i);
				res += model->L[index] + model->C[i] + (1.0 / model->b[index] + model->t[i]) * (M - model->M2);
			}
		}
	}
	return res;
}
