#ifndef PLOGP_H_
#define PLOGP_H_

#include "cpm_model.h"
#include <stdio.h>
#ifdef HAVE_CONFIG_H
#include "MPIBlib/benchmarks/mpib_measurement.h"
#else
#include "mpib_measurement.h"
#endif

/*!
 * \defgroup plogp The heterogeneous PLogP model
 * This module provides building of the heterogeneous extension of the parameterised LogP model
 * and PLogP/LogGP predictions of the execution time of
 * point-to-point and collective communication operations.
 * 
 * The PLogP model \latexonly\cite{Kielmann2000}\endlatexonly is an extension of
 * the LogP model \latexonly\cite{Culler1993}\endlatexonly.
 * The PLogP model is defined in terms of the following parameters:
 * - \f$ L \f$ - latency,
 * - \f$ o_s(M) \f$ and \f$ o_r(M) \f$ - sender and receiver overheads,
 * - \f$ g(M) \f$ - gap per message (delay between consecutive communications),
 * - \f$ P \f$ - number of processors.
 * 
 * Some of these parameters are functions, which makes this model "parameterised".
 * They are represented by piecewise linear functions. The PLogP parameters are estimated
 * with help of the logp_mpi library \latexonly\cite{Kielmann2000}\endlatexonly.
 * The heterogeneous extension is a set of the PLogP models built for each pair of processors:
 * \f$ \{L_{ij}, {o_s}_{ij}(M), {o_r}_{ij}(M), g_{ij}(M)\}_{i \ne j = 0}^{n-1} \f$.
 *  
 * The parameters of the LogP and LogGP (another extension of the LogP model 
 * proposed in \latexonly\cite{Alexandrov1995}\endlatexonly, which considers message size
 * and includes an extra parameter, \f$ G \f$, gap per byte)
 * models can be expressed via the PLogP parameters:
 * - \f$ L = L^p + g^p(1) - o_s^p(1) - o_r^p(1) \f$
 * - \f$ 2 * o = o_s^p(1) + o_r^p(1) \f$
 * - \f$ g = g^p(1) \f$
 * - \f$ G = g^p(M_{max}) / M_{max} \f$
 * 
 * The heterogeneous extension of the LogGP model can be obtained by applying the above equations
 * to the PLogP parameters of each pair of processors. Therefore, this module provides both
 * the PLogP and LogGP predictions of the execution time of collective communication operations.
 * \{
 */
#ifdef __cplusplus
extern "C" {
#endif

/*! The heterogeneous PLogP model. The point-to-point parameters for all pairs of processors. */
typedef struct PLogP_model {
	/*! Predictor */
	CPM_predictor predictor;
	/*! LogGP predictor */
	CPM_predictor LogGP_predictor;
	/*! Number of nodes */
	int n;
	/*!
	 * Array of \f$ C_n^2 \f$ pointers to the PLogP parameters, which correspond to all pairs of processors.
	 * Each pointer refers to the logp_params data structure of the logp_mpi library.
	 */
	void** logp;
} PLogP_model;

/*! Allocates memory for PLogP model */
PLogP_model* PLogP_alloc(int n);

/*! Frees the PLogP model. */
void PLogP_free(PLogP_model* model);

/*!
 * Reads the PLogP model.
 * \note Creates a temporary file.
 */
void PLogP_read(FILE* stream, PLogP_model** model);

/*!
 * Writes the PLogP model.
 * \note Creates a temporary file.
 */
void PLogP_write(FILE* stream, const PLogP_model* model);

/*!
 * Estimates the PLogP model. Calls the logp_mpi library.
 * \param comm communicator
 * \param precision measurement precision
 * \param msgset message set
 * \param parallel several non-overlapped point-to-point communications at the same time if non-zero
 * \param model PLogP model
 */
void PLogP_estimate(MPI_Comm comm, MPIB_precision precision, MPIB_msgset msgset, int parallel,
	PLogP_model** model);

/*! Predicts the execution time of a point-to-point communication as \f$ L + g(M)\f$ */
double PLogP_predict_p2p(void* _this, int i, int j, int M);

/*!
 * Predicts the execution time of a point-to-point communication according to LogGP model
 * as \f$ L + 2 * o + G (M - 1) \f$\n
 * \param _this PLogP model
 * \param i index of the process
 * \param j index of the process
 * \param M message size
 */
double LogGP_predict_p2p(void* _this, int i, int j, int M);

/*!
 * Homogeneous PLogP prediction of linear scatter/gather:
 * \f$ L + (n - 1) g(M)\f$
 */
double PLogP_hpredict_sg_linear(void* _this, int M);

/*!
 * Homogeneous LogGP prediction of linear scatter/gather:
 * \f$ L + 2 o + (n - 1) (M - 1) G + (n - 2) g \f$
 */
double LogGP_hpredict_sg_linear(void* _this, int M);

#ifdef __cplusplus
}
#endif
/*!
 * \}
 */

#endif /*PLOGP_H_*/
