/*!
 * \page pgemm Parallel matrix multiplication
 * Heterogeneous parallel matrix product:
 * -# Estimates the execution time of gsl_blas_dgemm on all processors
 * -# Scatters the row blocks of matrix A proportionally to the speed of processors (scatterv)
 * -# Broadcasts matrix B
 * -# Gathers matrix C (gatherv)
 *
 * Implemented with the following collective communications:
 * - native
 * - Hockney_dfs_binomial_min
 * - Hockney_dfs_binomial_max
 */

#include "config.h"
#include "models/cpm.h"
#include <string.h>
#include <getopt.h>
#include <gsl/gsl_blas.h>
#include "collectives/cpm_coll.h"

int main(int argc, char** argv) {
	MPI_Init(&argc, &argv);
	MPI_Comm comm = MPI_COMM_WORLD;
#ifndef CPM_DEBUG
	MPIB_Comm(MPI_COMM_WORLD, &comm);
	if (comm == MPI_COMM_NULL)
		goto exit;
#endif

	int exit;
	MPIB_getopt_help_default(&exit);
	int N = 1000;
	char operations[256] = "native";
	MPIB_Scatterv scatterv = MPI_Scatterv;
	MPIB_Bcast bcast = MPI_Bcast;
	MPIB_Gatherv gatherv = MPI_Gatherv;

	int rank;
	MPI_Comm_rank(comm, &rank);
	if (rank == 0)
	{
		int c;
		char options[256] = "N:O:";
		strcat(options, MPIB_getopt_help_options());
		while ((c = getopt(argc, argv, options)) != -1)
		{
			MPIB_getopt_help_optarg(c, "pgemm", &exit);
			switch (c)
			{
				case 'h':
					fprintf
					(
						stderr,
"-N I			matrix size (default: %d)\n\
-O S			collective operations: native, Hockney_dfs_binomial_min, Hockney_dfs_binomial_max (default: native)\n",
						N
					);
					break;
				case 'N':
					N = atoi(optarg);
					break;
				case 'O':
					strcpy(operations, optarg);
					break;
			}
		}
	}
	MPIB_getopt_help_bcast(&exit, 0, comm);
	if (exit)
		goto exit;

	MPI_Bcast(&N, 1, MPI_INT, 0, comm);
	int len;
	if (rank == 0)
		len = strlen(operations) + 1;
	MPI_Bcast(&len, 1, MPI_INT, 0, comm);
	MPI_Bcast(operations, len, MPI_CHAR, 0, comm);
	if (!strcmp(operations, "Hockney_dfs_binomial_min"))
	{
		scatterv = Hockney_Scatterv_dfs_binomial_min;
		bcast = Hockney_Bcast_dfs_binomial_min;
		gatherv = Hockney_Gatherv_dfs_binomial_min;
	}
	if (!strcmp(operations, "Hockney_dfs_binomial_max"))
	{
		scatterv = Hockney_Scatterv_dfs_binomial_max;
		bcast = Hockney_Bcast_dfs_binomial_max;
		gatherv = Hockney_Gatherv_dfs_binomial_max;
	}

	MPIB_print_processors(comm);
	if (rank == 0)
		printf("#pgemm A(%d, %d) * B(%d, %d) with %s scatterv, bcast, gatherv\n", N, N, N, N, operations);

	int m = 100;
	gsl_matrix* A = gsl_matrix_alloc(m, m);
	gsl_matrix_set_all(A, 1);
	gsl_matrix* B = gsl_matrix_alloc(m, m);
	gsl_matrix_set_all(B, 2);
	gsl_matrix* C = gsl_matrix_alloc(m, m);
	double speed = MPI_Wtime();
	gsl_blas_dgemm(CblasNoTrans, CblasNoTrans, 1, A, B, 0, C);
	speed = 1 / (MPI_Wtime() - speed);
	gsl_matrix_free(A);
	gsl_matrix_free(B);
	gsl_matrix_free(C);

	int size;
	MPI_Comm_size(comm, &size);
	double* speeds = rank == 0 ? (double*)malloc(sizeof(double) * size) : NULL;
	MPI_Gather(&speed, 1, MPI_DOUBLE, speeds, 1, MPI_DOUBLE, 0, comm);
	int* blocks = rank == 0 ? (int*)malloc(sizeof(int) * size) : NULL;
	int* counts = rank == 0 ? (int*)malloc(sizeof(int) * size) : NULL;
	int* displs = rank == 0 ? (int*)malloc(sizeof(int) * size) : NULL;
	if (rank == 0)
	{
		double s = 0;
		int i;
		for (i = 0; i < size; i++)
			s += speeds[i];
		int b = 0;
		for (i = 0; i < size - 1; i++)
			b += blocks[i] = N * speeds[i] / s;
		blocks[size - 1] = N - b;
		printf("#rank\tspeed\t\tblock\n");
		for (i = 0; i < size; i++)
		{
			counts[i] = blocks[i] * N;
			displs[i] = i == 0 ? 0 : displs[i - 1] + counts[i - 1];
			printf("#%d\t%le\t%d\n", i, speeds[i], blocks[i]);
		}
	}
	free(speeds);
	int block;
	MPI_Scatter(blocks, 1, MPI_INT, &block, 1, MPI_INT, 0, comm);
	free(blocks);

	gsl_matrix* matrix = rank == 0 ? gsl_matrix_alloc(N, N) : NULL;
	if (rank == 0)
		gsl_matrix_set_all(matrix, 1);
	A = gsl_matrix_alloc(block, N);
	B = gsl_matrix_alloc(N, N);
	gsl_matrix_set_all(B, 2);
	C = gsl_matrix_alloc(block, N);
	double time = MPI_Wtime();
	scatterv(rank == 0 ? matrix->data : NULL, counts, displs, MPI_DOUBLE, A->data, block * N, MPI_DOUBLE, 0, comm);
	if (rank == 0)
		printf("#scatterv: %le\n", MPI_Wtime() - time);
	time = MPI_Wtime();
	bcast(B->data, N * N, MPI_DOUBLE, 0, comm);
	if (rank == 0)
		printf("#bcast: %le\n", MPI_Wtime() - time);
	time = MPI_Wtime();
	gsl_blas_dgemm(CblasNoTrans, CblasNoTrans, 1, A, B, 0, C);
	if (rank == 0)
		printf("#gemm: %le\n", MPI_Wtime() - time);
	time = MPI_Wtime();
	gatherv(C->data, block * N, MPI_DOUBLE, rank == 0 ? matrix->data : NULL, counts, displs, MPI_DOUBLE, 0, comm);
	if (rank == 0)
		printf("#gatherv: %le\n", MPI_Wtime() - time);
	if (rank == 0)
		gsl_matrix_free(matrix);
	gsl_matrix_free(A);
	gsl_matrix_free(B);
	gsl_matrix_free(C);

	free(counts);
	free(displs);

exit:
#ifndef CPM_DEBUG
	MPI_Comm_free(&comm);
#endif
	MPI_Finalize();
	return 0;
}
