/*!
 * \page generate_hierarchy_input_file Generator of hierarchy files
 * This tool generates a hierarchy file automatically. At the moment, it bases its hierarchy on following rule: 
 * - if hostnames start with the same 3 characters, they are clustered under the same parent
 * - after that, nodes are clustered according to the domain name after the first "."
 * It is assumed that the environment will return the fully qualified name. The tool can be extended easily to work further on the fully qualified name

 * Usage: 
 * \code
 * $ mpirun [mpi options] generate_hierarchy_input_file <prefix for cluster 1> <prefix for cluster 2>  ... >  h-file.out
 * \endcode
 *
 * The output of this tool can be used as input for the \ref hierarchical_model. The output file might need
 * some manual modifications.
 *
 * For visualizing the ranks on each host, see \ref ranks2hosts
 */
#include "config.h"
#include <boost/tokenizer.hpp>
#include <boost/algorithm/string/predicate.hpp>
#include "models/hierarchical/hierarchical_model.hpp"

int main(int argc, char **argv) {
	mpi::environment env(argc, argv);
	mpi::communicator world;
	std::map<string, Vertex> name2vertex;

	if (argc < 2) {
		cerr << "User must specify a list of common prefixes for the nodes on each cluster\n" << endl;
		exit(-1);
	}
	std::vector<std::string> args;
	for (int i=1; i< argc; i++) 
		args.push_back(argv[i]);

	string proc_name = env.processor_name();
	if (world.rank() != 0) {
		gather(world, proc_name, 0);
	}
	else {
		std::vector<std::string> all_proc_names;
		gather(world, proc_name, all_proc_names, 0);	
		Graph graph;
		//this creates one vertex per hostname, no repetitions allowed
		for (std::vector<string>::iterator it =	all_proc_names.begin(); it != all_proc_names.end(); it++) {
			string str = *it;
			if (name2vertex.find(str) == name2vertex.end()) {
				//a leaf node is created
				Vertex v1 = boost::add_vertex(graph);
				put(vertex_model_data_t(), graph, v1, str); 
				put(vertex_type_t(), graph, v1, 0); 
				name2vertex[str] = v1;
				//adjacent node for on-core properties
				Vertex v2 = boost::add_vertex(graph);
				put(vertex_type_t(), graph, v2, 1); 
				boost::add_edge(v1, v2, graph);
				
				//Look for the prefix common to all nodes on a cluster - based on user input
				std::string start_str("");
				for (std::vector<std::string>::iterator it = args.begin(); it != args.end(); it++) 
					if (boost::starts_with(str, *it))
						start_str = *it;
				if (!start_str.compare("")) {
					//start_str = *it;
					cerr << "User did not provide all possible occurring prefixes!\n";
					exit(-1);
				}
				if (name2vertex.find(start_str) == name2vertex.end()) {
					Vertex v3 = boost::add_vertex(graph);	
					put(vertex_type_t(), graph, v3, 1); 
					name2vertex[start_str] = v3;
					boost::char_separator<char> sep(".");
					boost::tokenizer<boost::char_separator<char> > tok(str, sep);
					boost::tokenizer<boost::char_separator<char> >::iterator tok_it = tok.begin();
					//check there is "."
					if ((tok_it != tok.end()) && (++tok_it != tok.end())) {
						string next_tok = *tok_it;
						//node representing the whole site
						if (name2vertex.find(next_tok) == name2vertex.end()) {
							Vertex v4 = boost::add_vertex(graph);	
							put(vertex_type_t(), graph, v4, 1); 
							name2vertex[next_tok] = v4;
						}
						boost::add_edge(v3, name2vertex[next_tok], graph);
					}
				}
				boost::add_edge(v2, name2vertex[start_str], graph);
			}
		}
		boost::write_graphviz(cout, graph, CPM::hierarchy::label_writer(graph));
	}
	return 0;
}

