/*!
 * \page hierarchical_model Hierarchical model builder
 *
 * This tool builds a model file efficiently based on hierarchy information. The input file should be a hierarchical graphviz file (DOT format)
 * with a single root element. Every element has two attributes:
 * 	- type (0 or 1) - to denote a leaf or non-leaf element
 *	- model_data (empty string or model-specific data/parameters/etc) - can be initially used
 *	for storing hostnames
 *
 * The output file itself is not hierarchical but flat.
 *	\b Example:
 * \code
 * $ mpirun -n 2 hierarchical_model -C Hockney -i h-file.out -o model.out
 * \endcode
 *
 * The model.out file can then be used by either of these tools:
 * - \ref prediction
 * - \subpage collective
 *
 */
#include "config.h"
#include "models/hierarchical/hierarchical_model.hpp"
#include "models/hierarchical/hockney_p2p_model.hpp"
#include "models/hierarchical/plogp_p2p_model.hpp"
#include "models/cpm.h"
#include <iostream>
#include <fstream>
#include <getopt.h>

using namespace std;

int main(int argc, char **argv) {
	MPI_Init(&argc, &argv);
	MPI_Comm comm = MPI_COMM_WORLD;
	int root = 0;

	int exit;
	MPIB_getopt_help_default(&exit);
	MPIB_msgset msgset;
	MPIB_getopt_msgset_default(&msgset);
	MPIB_precision precision;
	MPIB_getopt_precision_default(&precision);
	int parallel;
	int p2p_type, proc_num_spawned /* not used now */;
	ifstream is;
	MPIB_getopt_p2p_default(&parallel, &p2p_type, &proc_num_spawned);
	char model_name[256] = "";
	//FILE* ostream = stdout;
	ofstream os;
	int rank;
	MPI_Comm_rank(comm, &rank);
	int size;
	MPI_Comm_size(comm, &size);
	if (rank == 0) {
		int c;
		char options[256] = "C:i:o:";
		strcat(options, MPIB_getopt_help_options());
		strcat(options, MPIB_getopt_msgset_options());
		strcat(options, MPIB_getopt_precision_options());
		strcat(options, MPIB_getopt_p2p_options());
		while ((c = getopt(argc, argv, options)) != -1) {
			MPIB_getopt_help_optarg(c, "model", &exit);
			MPIB_getopt_precision_optarg(c, &precision);
			MPIB_getopt_msgset_optarg(c, &msgset);
			MPIB_getopt_p2p_optarg(c, &parallel, &p2p_type, &proc_num_spawned);
			switch (c) {
				case 'h':
				cerr <<	"\n\
-C S			communication model (required): Hockney, PLogP\n\
-i S			input file (optional)\n\
-o S			output file (default: stdout)\n";
					break;
				case 'C':
					strcpy(model_name, optarg);
					break;
				case 'i':
					is.open(optarg);
					break;
				case 'o':
					os.open(optarg);
					if (!os) {
						cerr << "Can't write into file\n";
						MPI_Abort(comm, -1);
					}
					break;
			}
		}
	}
	MPIB_getopt_help_bcast(&exit, 0, comm);
	if (exit) {
		MPI_Finalize();
		return 0;
	}
	
	MPIB_getopt_msgset_bcast(&msgset, 0, comm);
	MPIB_getopt_precision_bcast(&precision, 0, comm);
	MPIB_getopt_p2p_bcast(&parallel, &p2p_type, &proc_num_spawned, 0, comm);

	int len;
	len = strlen(model_name) + 1;
	MPI_Bcast(&len, 1, MPI_INT, 0, comm);
	MPI_Bcast(&model_name, len, MPI_CHAR, 0, comm);
	int build; 
	if (rank == 0)
		build = (is)?1:0;
	MPI_Bcast(&build, 1, MPI_INT, 0, comm);
	if (build)
	{
		double time;
		if (!strcmp(model_name, "Hockney")) {
			time = MPI_Wtime();
			time = MPI_Wtime() - time;
		}
		if (!strcmp(model_name, "PLogP")) {
			time = MPI_Wtime();
			time = MPI_Wtime() - time;
		}
		MPIB_print_processors(comm);
		if (rank == 0) {
			CPM_print_model(model_name, msgset, parallel, precision, time);
		//	if (!strcmp(model_name, "Hockney"))
//				Hockney_write(ostream, (Hockney_model*)model);
		}
	}
	else {
		if (!is) {
			if (rank == 0)
				cerr << "Cannot open model file for reading" << endl;
			MPI_Abort(comm, -1);
		}
	}

	if (!strcmp(model_name, "Hockney")) {
		H_Model<hockney_p2p_model>* model = new H_Model<hockney_p2p_model>;
		if (rank == root) {
			model->read_hierarchy_file(is);
			is.close();
		}
		H_Model_estimate(comm, msgset, precision, parallel, model);
		if (rank ==0) {
			model->write_model_file(os);
			//boost::write_graphviz(cout, model->getGraph(), label_writer(model->getGraph()));
			os.close();
			delete model;
		}
	}
	if (!strcmp(model_name, "PLogP")) {
		H_Model<plogp_p2p_model>* model = new H_Model<plogp_p2p_model>;
		if (rank == root) {
			model->read_hierarchy_file(is);
			is.close();
		}
		H_Model_estimate(comm, msgset, precision, parallel, model);
		if (rank ==0) {
			model->write_model_file(os);
			//boost::write_graphviz(cout, model->getGraph(), label_writer(model->getGraph()));
			os.close();
			delete model;
		}
	}

	MPI_Finalize();
	return 0;
}
