/*!
 * \page model Model builder
 * Builds the Hockney, PLogP and LMO models and estimates the execution time of point-to-point and
 * collective communication operations.
 * Arguments are described in the MPIBlib manual.
 * 
 * \b Example:
 * \code
 * $ mpirun -n 16 model -C Hockney -o model.out
 * \endcode
 *
 * The model.out file can then be used by either of these tools:
 * 	- \ref prediction
 * 	- \subpage collective
 *
 * \page collective Collective benchmark (using models)
 *
 * The basic usage of this tool and a description how to plot the output file is described in MPIBlib.
 *
 * \b Example using a model-based collective:
 * \code
 * $ mpirun -n 16 collective -O Hockney_Scatter_dfs_binomial_min -o model=Hockney,file=model.out -l libcpm_coll.so -m 1024 -M 2049 > collective.out
 * \endcode
 * will benchmark a binomial tree scatter operation with message sizes per process of 1024 and 2048. The binomial tree will
 * be generated through depth-first approach choosing the node with minimum predicted communication by
 * the Hockney model.
 *
 */

#include "config.h"

#include <string.h>
#include <getopt.h>
#include <gsl/gsl_statistics_double.h>

#include "MPIBlib/p2p/mpib_p2p.h"
#include "models/cpm.h"
#ifdef HAVE_LIBR
#include "models/cpm_R.h"
#endif /*HAVE_LIBR*/

int main(int argc, char** argv) {
	MPI_Init(&argc, &argv);
	MPI_Comm comm = MPI_COMM_WORLD;
#ifndef CPM_DEBUG
	MPIB_Comm(MPI_COMM_WORLD, &comm);
	if (comm == MPI_COMM_NULL) {
		MPI_Finalize();
		return 0;
	}
#endif

	int exit;
	MPIB_getopt_help_default(&exit);
	MPIB_msgset msgset;
	MPIB_getopt_msgset_default(&msgset);
	MPIB_precision precision;
	MPIB_getopt_precision_default(&precision);
	int parallel;
	int p2p_type, proc_num_spawned;
	MPIB_getopt_p2p_default(&parallel, &p2p_type, &proc_num_spawned);
	char model_name[256] = "";
	void* model = NULL;
	FILE* ostream = stdout;

	int rank;
	MPI_Comm_rank(comm, &rank);
	int size;
	MPI_Comm_size(comm, &size);
	if (rank == 0) {
		int c;
		char options[256] = "C:o:";
		strcat(options, MPIB_getopt_help_options());
		strcat(options, MPIB_getopt_msgset_options());
		strcat(options, MPIB_getopt_precision_options());
		strcat(options, MPIB_getopt_p2p_options());
		while ((c = getopt(argc, argv, options)) != -1) {
			MPIB_getopt_help_optarg(c, "model", &exit);
			MPIB_getopt_precision_optarg(c, &precision);
			MPIB_getopt_msgset_optarg(c, &msgset);
			MPIB_getopt_p2p_optarg(c, &parallel, &p2p_type, &proc_num_spawned);
			switch (c) {
				case 'h':
					fprintf(
						stderr,
"-C S			communication model (required): Hockney, PLogP, LMO\n\
-o S			output file (default: stdout)\n"
					);
					break;
				case 'C':
					strcpy(model_name, optarg);
					break;
				case 'o':
					ostream = fopen(optarg, "w");
					if (ostream == NULL)
						fprintf(stderr, "Cannot open file %s for writing\n", optarg);
					break;
			}
		}
	}
	MPIB_getopt_help_bcast(&exit, 0, comm);
	if (exit) {
		MPI_Finalize();
		return 0;
	}
	
	MPIB_getopt_msgset_bcast(&msgset, 0, comm);
	MPIB_getopt_precision_bcast(&precision, 0, comm);
	MPIB_getopt_p2p_bcast(&parallel, &p2p_type, &proc_num_spawned, 0, comm);

	int len;
	len = strlen(model_name) + 1;
	MPI_Bcast(&len, 1, MPI_INT, 0, comm);
	MPI_Bcast(&model_name, len, MPI_CHAR, 0, comm);
	
	if (p2p_type == 1)
		p2p_init(comm, proc_num_spawned);

	double time;
	if (!strcmp(model_name, "Hockney")) {
		time = MPI_Wtime();
		Hockney_estimate_regression(comm, msgset, precision, parallel, (Hockney_model**)&model);
		//Hockney_estimate(comm, msgset.max_size, precision, parallel, (Hockney_model**)&model);
		time = MPI_Wtime() - time;
	}
	if (!strcmp(model_name, "LinInterp")) {
		time = MPI_Wtime();
		LinInterp_estimate(comm, precision,msgset, parallel, (LinInterp_model**)&model);
		time = MPI_Wtime() - time;
	}
	if (!strcmp(model_name, "LMO")) {
#ifdef HAVE_LIBR
		if (rank == 0) {
			exit = CPM_initR();
			if (exit)
				fprintf(stderr, "R cannot be initialized\n");
		}
		MPI_Bcast(&exit, 1, MPI_INT, 0, comm);
		if (exit) {
			MPI_Finalize();
			return 0;
		}
#endif /*HAVE_LIBR*/
		time = MPI_Wtime();
		LMO_estimate(comm, precision, msgset, parallel, (LMO_model**)&model);
		time = MPI_Wtime() - time;
#ifdef HAVE_LIBR
		if (rank == 0)
			CPM_endR();
#endif /*HAVE_LIBR*/
	}
	if (!strcmp(model_name, "PLogP")) {
		time = MPI_Wtime();
		PLogP_estimate(comm, precision, msgset, parallel, (PLogP_model**)&model);
		time = MPI_Wtime() - time;
	}
	if (rank == 0) {
		exit = model == NULL;
		if (exit)
			fprintf(stderr, "Undefined model: %s\n", model_name);
	}
	MPI_Bcast(&exit, 1, MPI_INT, 0, comm);
	if (exit) {
		MPI_Finalize();
		return 0;
	}
	MPIB_print_processors(comm);
	if (rank == 0) {
		CPM_print_model(model_name, msgset, parallel, precision, time);
		if (!strcmp(model_name, "Hockney"))
			Hockney_write(ostream, (Hockney_model*)model);
		if (!strcmp(model_name, "LinInterp"))
			LinInterp_write(ostream, (LinInterp_model*)model);
		if (!strcmp(model_name, "PLogP"))
			PLogP_write(ostream, (PLogP_model*)model);
		if (!strcmp(model_name, "LMO"))
			LMO_write(ostream, (LMO_model*)model);
	}

	if (!strcmp(model_name, "Hockney"))
		Hockney_free((Hockney_model*)model);
	if (!strcmp(model_name, "LinInterp"))
		LinInterp_free((LinInterp_model*)model);
	if (!strcmp(model_name, "PLogP"))
		PLogP_free((PLogP_model*)model);
	if (!strcmp(model_name, "LMO"))
		LMO_free((LMO_model*)model);

	if (rank == 0 && ostream != stdout)
		fclose(ostream);

#ifndef CPM_DEBUG
	MPI_Comm_free(&comm);
#endif

	if (p2p_type == 1) 
		p2p_finalize();
	MPI_Finalize();
	return 0;
}
