/*!
 * \page prediction Predictor of the execution time of p2p or collective communication
 * Instead of performing p2p communication or model-based collective communication, this
 * routine performs the model-based prediction how long the communication would take. This tool
 * does not require a parallel MPI run.
 *
 * \b Examples:
 * \code
 * $ prediction -n <np> -O p2p -i model.out -C Hockney -m 1024 -M 2049
 * \endcode
 * will print the predicted P2P communication time for message sizes 1024 and 2048 bytes
 * for each of the \<np\> processes with each other process. It is expected
 * that the model.out was generated for \<np\> processes as well.
 *
 * \code
 * $ prediction -n <np> -O MPIB_Scatter_binomial -i model.out -C Hockney -m 1024 -M 2049
 * \endcode
 * will print the predicted communication time for the binomial tree implementation of MPI_Scatter
 * with message size of 1024 and 2048 bytes processes. It is expected
 * that the model.out was generated for \<np\> processes as well.
 *
 */
#include "config.h"

#include <getopt.h>
#include <malloc.h>
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <ctype.h>

#include "models/cpm.h"

int main(int argc, char** argv) {
	void* model = NULL;
	char operation[256]="";
	int np;
	char model_name[256]="";
	char options[256]="hi:m:M:S:n:f:O:f:C:";
	int c;
	double* factors;
	int i;
	FILE* file;
	char factors_filename[256] = "";
	MPIB_msgset msgset;
	MPIB_getopt_msgset_default(&msgset);
	CPM_coll_ops op;
	while ((c = getopt (argc, argv, options)) != -1) {
		switch (c) {
		case 'C':
			strcpy(model_name, optarg);
			break;
		case 'm':
			msgset.min_size = atoi(optarg);
			break;
		case 'f': 
			strcpy(factors_filename, optarg);
			break;
		case 'M':
			msgset.max_size = atoi(optarg);
			break;
		case 'S':
			msgset.stride = atoi(optarg);
			break;
		case 'i': 
			file = fopen(optarg, "r");
			if (file == NULL) {
				fprintf(stderr, "Cannot read input file %s\n", optarg);
				exit(-1);
			}
			break;
		case 'O': 
			strcpy(operation, optarg);
			for (i=0; i<256; i++) {
				operation[i] = toupper(operation[i]);
			}	
			if (-1 == ( op = findOperation(operation)))  {
				fprintf(stderr, "could not find prediction for this operation\n");
				exit(-1);
			}
			break;
		case 'n':
			np = atoi(optarg);
			break;
		//The help is glued together from the prediction options and a subset of the options in MPIBlib getopt
		case 'h':
			fprintf(
				stderr,
"-v			verbose information on the process of building trees\n\
-n 			number of processes (= number of communication tree nodes) (required)\n\
-O 			'p2p' for P2P predictions for all pairs or name of collective operation to predict (required)\n\
-i 			model input file (required)\n\
-f 			input file containing the multiplying factors for scatterv/gatherv (default: all factors = 1) (optional)\n\
-m 			minimum message size (default: %d)\n\
-M 			maximum message size (default: %d)\n\
-S 			stride between message sizes (default: %d)\n\
-C 			model name (this option is only relevant for predicting MPIB_* operations)\n ",
			msgset.min_size,
			msgset.max_size,
			msgset.stride);
			exit(0);
		default:
			
			fprintf(stderr, "Argument: %c\n", c);
			exit(-1);
		}
	}//finished with user arguments

	//printf("DEBUG:-1\n");
	CPM_model model_type = NULL_MODEL;
	model_type = getModel(findOperation(operation));
	if (model_type == HOCKNEY) 
		Hockney_read(file, (Hockney_model**) &model);
	if (model_type == LININTERP) {
		LinInterp_read(file, (LinInterp_model**) &model);
	}
	if(model_type == PLOGP || model_type == LOGGP){
		PLogP_read(file, (PLogP_model**)&model);
	}
#ifdef HAVE_LIBR
	if (model_type == LMO) {
		LMO_read(file, (LMO_model**) &model);
	}
#endif
	if ((model_type == MPIB_) || (model_type == NULL_MODEL && (0 == strcmp(operation, "P2P")))) {
		if (!strcmp(model_name, "Hockney") || !strcmp(model_name, ""))
			Hockney_read(file, (Hockney_model**) &model);
		else if (!strcmp(model_name, "PLogP") || !strcmp(model_name, "LogGP"))
			PLogP_read(file, (PLogP_model**) &model);
		else if (!strcmp(model_name, "LinInterp"))
			LinInterp_read(file, (LinInterp_model**) &model);
#ifdef HAVE_LIBR
		else if (!strcmp(model_name, "LMO"))
			LMO_read(file, (LMO_model**) &model);
#endif
		else {
			fprintf(stderr, "User specified unknown model\n");
			exit (-1);
		}
	}
	if (model == NULL) {
		fprintf(stderr, "Model is NULL\n");
		exit (-1);
	}
	int M;
	MPIB_result result;
	//P2P predictions
	if (!strcmp("P2P", operation)) {
		//P2P predictions for the set of message sizes
		CPM_predictor* predictor;
		if (!strcmp(model_name, "Hockney") || !strcmp(model_name, ""))
			predictor = &((Hockney_model *)model)->predictor;
		else if (!strcmp(model_name, "PLogP"))
			predictor = &((PLogP_model *)model)->predictor;
		else if (!strcmp(model_name, "LinInterp"))
			predictor = &((LinInterp_model *)model)->predictor;
		else if (!strcmp(model_name, "LogGP"))
			predictor = &((PLogP_model *)model)->LogGP_predictor;
#ifdef HAVE_LIBR
		else if (!strcmp(model_name, "LMO"))
			predictor = &((LMO_model *)model)->predictor;
#endif
		else {
			fprintf(stderr, "User specified unknown model\n");
			exit (-1);
		}
		MPIB_pairs* pairs = MPIB_build_pairs(np);
		MPIB_pairs* iter_pairs = pairs;
		while (iter_pairs) {
			MPIB_pair* iter_pair = iter_pairs->list;
			while (iter_pair) {
				int i = iter_pair->values[0];
				int j = iter_pair->values[1];
				if (i != j) {
					printf("P2P predictions for pair (%d-%d)\n", i, j);
					for (M = msgset.min_size; M < msgset.max_size; M += msgset.stride) {
						result.T = predictor->predict_p2p((!strcmp(model_name,"LogGP")? predictor:model), i, j, M);
						result.wtick = 0.;
						result.reps = 0;
						result.ci = 0;
						MPIB_print_coll_tr(M, result);
					}
				}
				iter_pair = iter_pair->next;
			}
			iter_pairs = iter_pairs->next;
		}
	}
	//Predictions for collective operations except for SGv
	else if (op < MPIB_SCATTERV_LINEAR) {
		//predictions for the set of message sizes
		for (M = msgset.min_size; M < msgset.max_size; M += msgset.stride) {
			result.T = CPM_predict_brsg(model, np, 0, M, op);
			result.wtick = 0.;
			result.reps = 0;
			result.ci = 0;
			MPIB_print_coll_tr(M, result);
		}
	}
	//Predictions for SGv operations
	else {
		MPIB_result result;
		factors = (double *) malloc (sizeof(double) * np);
		//user specified a factors file
		if (strcmp(factors_filename, "")) {
			FILE* stream = fopen(factors_filename, "r");
			if (stream == NULL) {
				fprintf(stderr, "Cannot open file %s for reading\n", optarg);
				exit(-1) ;
			} else {
				int i;
				for (i = 0; i < np; i++)
					fscanf(stream, "%le\n", &factors[i]);
				fclose(stream);
			}
		}
		//No factors file specified - default factors to 1's
		else {
			int i;
			for (i = 0; i < np; i++)
				factors[i]=1.;
		}
		//determine the message sizes
		for (M = msgset.min_size; M < msgset.max_size; M += msgset.stride) {
			int i;
			int* counts = (int *) malloc (sizeof(int) * np);
			for (i=0; i < np; i++) {
				counts[i] = M * factors[i];
			}
			result.T = CPM_predict_sgv(model, np, 0, counts, op);
			free(counts);
			result.wtick = 0.;
			result.reps = 0;
			result.ci = 0;
			MPIB_print_coll_tr(M, result);
		}
	}

	/*
	 * free the models finally
	 */
	if (model_type == HOCKNEY)
			Hockney_free((Hockney_model*) model);
		if (model_type == LININTERP) {
			LinInterp_free((LinInterp_model*) model);
		}
		if(model_type == PLOGP || model_type == LOGGP){
			PLogP_free((PLogP_model*)model);
		}
	#ifdef HAVE_LIBR
		if (model_type == LMO) {
			LMO_free((LMO_model*) model);
		}
	#endif
		if ((model_type == MPIB_) || (model_type == NULL_MODEL && (0 == strcmp(operation, "P2P")))) {
			if (!strcmp(model_name, "Hockney") || !strcmp(model_name, ""))
				Hockney_free((Hockney_model*) model);
			else if (!strcmp(model_name, "PLogP") || !strcmp(model_name, "LogGP"))
				PLogP_free((PLogP_model*) model);
			else if (!strcmp(model_name, "LinInterp"))
				LinInterp_free((LinInterp_model*) model);
	#ifdef HAVE_LIBR
			else if (!strcmp(model_name, "LMO"))
				LMO_free( (LMO_model*) model);
	#endif
			else {
				fprintf(stderr, "User specified unknown model\n");
				exit (-1);
			}
		}
	return 0;
}
